#!/usr/bin/env python
import logging
import os
import json
import sys
from datetime import datetime
import inspect

# from rich.prompt import Confirm
import click
import relecov_tools.config_json
import relecov_tools.download_manager
import relecov_tools.log_summary
import rich.console
import rich.traceback

import relecov_tools.config_json
import relecov_tools.utils
import relecov_tools.read_lab_metadata
import relecov_tools.download_manager
import relecov_tools.json_validation
import relecov_tools.mail
import relecov_tools.map_schema
import relecov_tools.upload_database
import relecov_tools.read_bioinfo_metadata
import relecov_tools.metadata_homogeneizer
import relecov_tools.gisaid_upload
import relecov_tools.upload_ena_protocol
import relecov_tools.pipeline_manager
import relecov_tools.build_schema
import relecov_tools.dataprocess_wrapper
import relecov_tools.upload_results
import relecov_tools.base_module

log = logging.getLogger()

# Set up rich stderr console
stderr = rich.console.Console(
    stderr=True, force_terminal=relecov_tools.utils.rich_force_colors()
)

__version__ = "1.5.5"

# IMPORTANT: When defining a Click command function in this script,
# you MUST include both 'ctx' (for @click.pass_context) and ALL the parameters
# defined by @click.option decorators in the function signature.
# Example:
# @click.pass_context
# def my_command(ctx, param1, param2, ...):
#     ...
# This is required for correct argument passing and to avoid runtime errors.


# Set up  merge config with extra plus CLI
def merge_with_extra_config(ctx, add_extra_config=False):
    """
    Merge CLI arguments with those defined in config or extra_config.
    CLI has a higher priority than extra_config, which has a higher priority than default (None).
    Additionally, convert any empty string values ('') to None.
    """
    # Set which configuration is going to be used
    if add_extra_config:
        config = relecov_tools.config_json.ConfigJson(extra_config=True)
    else:
        config = relecov_tools.config_json.ConfigJson()
    ctx.obj["config"] = config.json_data

    command_name = ctx.command.name.replace("-", "_")
    command_params = ctx.params
    extra_args = config.json_data.get(command_name, {})

    # Merge: CLI > extra_config > default
    merged = dict(extra_args)
    for k, v in command_params.items():
        if v is not None:
            merged[k] = v

    # Convert empty strings to None
    for k, v in merged.items():
        if v == "":
            merged[k] = None

    # Get function signature to filter only valid params
    func = ctx.command.callback
    sig = inspect.signature(func)
    param_names = list(sig.parameters.keys())

    filtered = {k: v for k, v in merged.items() if k in param_names}
    return filtered


def run_relecov_tools():
    # Set up the rich traceback
    rich.traceback.install(console=stderr, width=200, word_wrap=True, extra_lines=1)

    # Print nf-core header
    # stderr.print("\n[green]{},--.[grey39]/[green],-.".format(" " * 42), highlight=False)
    stderr.print(
        "[blue]                ___   ___       ___  ___  ___                           ",
        highlight=False,
    )
    stderr.print(
        "[blue]   \    |-[grey39]-|  [blue] |   \ |    |    |    |    |   | \      /  ",
        highlight=False,
    )
    stderr.print(
        "[blue]    \   \  [grey39]/ [blue]  |__ / |__  |    |___ |    |   |  \    /   ",
        highlight=False,
    )
    stderr.print(
        "[blue]    /  [grey39] / [blue] \   |  \  |    |    |    |    |   |   \  /    ",
        highlight=False,
    )
    stderr.print(
        "[blue]   /   [grey39] |-[blue]-|   |   \ |___ |___ |___ |___ |___|    \/     ",
        highlight=False,
    )

    # stderr.print("[green]                                          `._,._,'\n", highlight=False)
    stderr.print(
        "\n" "[grey39]    RELECOV-tools version {}".format(__version__), highlight=False
    )

    # Lanch the click cli
    relecov_tools_cli()


# Customise the order of subcommands for --help
class CustomHelpOrder(click.Group):
    def __init__(self, *args, **kwargs):
        self.help_priorities = {}
        super(CustomHelpOrder, self).__init__(*args, **kwargs)

    def get_help(self, ctx):
        self.list_commands = self.list_commands_for_help
        return super(CustomHelpOrder, self).get_help(ctx)

    def list_commands_for_help(self, ctx):
        """reorder the list of commands when listing the help"""
        commands = super(CustomHelpOrder, self).list_commands(ctx)
        return (
            c[1]
            for c in sorted(
                (self.help_priorities.get(command, 1000), command)
                for command in commands
            )
        )

    def command(self, *args, **kwargs):
        """Behaves the same as `click.Group.command()` except capture
        a priority for listing command names in help.
        """
        help_priority = kwargs.pop("help_priority", 1000)
        help_priorities = self.help_priorities

        def decorator(f):
            cmd = super(CustomHelpOrder, self).command(*args, **kwargs)(f)
            help_priorities[cmd.name] = help_priority
            return cmd

        return decorator


@click.group(cls=CustomHelpOrder)
@click.version_option(relecov_tools.__version__)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Print verbose output to the console.",
)
@click.option(
    "-l",
    "--log-path",
    default=None,
    help="Creates log file in given folder. Uses default path in config or tmp if empty.",
)
@click.option(
    "-d",
    "--debug",
    is_flag=True,
    default=False,
    help="Show the full traceback on error for debugging purposes.",
)
@click.option(
    "-h",
    "--hex-code",
    default=None,
    help="Define hexadecimal code. This might overwrite existing files with the same hex-code",
)
@click.pass_context
def relecov_tools_cli(ctx, verbose, log_path, debug, hex_code):
    if debug:
        # Set the base logger to output everything
        level = logging.DEBUG
    else:
        # Set the base logger to hide DEBUG messages
        level = logging.INFO
    log.setLevel(level)
    if verbose:
        stream_handler = relecov_tools.base_module.BaseModule.set_log_handler(
            None, level=level, only_stream=True
        )
        log.addHandler(stream_handler)
    if hex_code is not None:
        relecov_tools.base_module.BaseModule._global_hex_code = str(hex_code)
    # Set up logs to a file if we asked for one
    called_module = str(ctx.invoked_subcommand).replace("-", "_")
    if os.path.isfile(relecov_tools.config_json.ConfigJson._extra_config_path):
        config = relecov_tools.config_json.ConfigJson(extra_config=True)
    else:
        config = relecov_tools.config_json.ConfigJson()
    logs_config = config.get_configuration("logs_config")
    default_outpath = logs_config.get("default_outpath", "/tmp/relecov_tools")
    if log_path is None:
        log_path = logs_config.get("modules_outpath", {}).get(called_module)
        if not log_path:
            log_path = os.path.join(default_outpath, called_module)
    else:
        relecov_tools.base_module.BaseModule._cli_log_path_param = log_path
    current_datetime = datetime.today().strftime("%Y%m%d%-H%M%S")
    log_filepath = os.path.join(
        log_path, "_".join([called_module, current_datetime]) + ".log"
    )

    try:
        os.makedirs(log_path, exist_ok=True)
        log_fh = relecov_tools.base_module.BaseModule.set_log_handler(
            log_filepath, level=level
        )
        log.addHandler(log_fh)
    except Exception:
        log_filepath = os.path.join(
            default_outpath, "_".join([called_module, "temp"]) + ".log"
        )
        os.makedirs(default_outpath, exist_ok=True)
        log_fh = relecov_tools.base_module.BaseModule.set_log_handler(
            log_filepath, level=level
        )
        log.addHandler(log_fh)
        log.warning(f"Invalid --log-path {log_path}. Using {log_filepath} instead")
    relecov_tools.base_module.BaseModule._cli_log_file = os.path.realpath(log_filepath)
    cli_command = " ".join(sys.argv)
    relecov_tools.base_module.BaseModule._cli_command = cli_command
    relecov_tools.base_module.BaseModule._current_version = __version__
    ctx.ensure_object(dict)  # Asegura que ctx.obj es un diccionario
    ctx.obj["debug"] = debug  # Guarda el flag de debug


# sftp
@relecov_tools_cli.command(help_priority=2)
@click.option("-u", "--user", help="User name for login to sftp server")
@click.option("-p", "--password", help="password for the user to login")
@click.option(
    "-f",
    "--conf_file",
    help="Configuration file (not params file)",
)
@click.option(
    "-d",
    "--download_option",
    default=None,
    multiple=False,
    help="Select the download option: [download_only, download_clean, delete_only]. \
        download_only will only download the files \
        download_clean will remove files from sftp after download \
        delete_only will only delete the files",
)
@click.option(
    "-o",
    "--output_dir",
    "--output-dir",
    "--output_folder",
    "--out-folder",
    "--output_location",
    "--output_path",
    "--out_dir",
    "--output",
    "output_dir",
    type=click.Path(file_okay=False, resolve_path=True),
    help="Directory where the generated output will be saved",
)
@click.option(
    "-t",
    "--target_folders",
    is_flag=False,
    flag_value="ALL",
    default=None,
    help='Flag: Select which folders will be targeted giving [paths] or via prompt. For multiple folders use ["folder1", "folder2"]',
)
@click.option(
    "-s",
    "--subfolder",
    is_flag=False,
    flag_value="ALL",
    default="RELECOV",
    help="Flag: Specify which subfolder to process (default: RELECOV)",
)
@click.pass_context
def download(
    ctx,
    user,
    password,
    conf_file,
    download_option,
    output_dir,
    target_folders,
    subfolder,
):
    """Download files located in sftp server."""
    debug = ctx.obj.get("debug", False)
    args_merged = merge_with_extra_config(
        ctx=ctx,
        add_extra_config=True,
    )
    try:
        download_manager = relecov_tools.download_manager.DownloadManager(**args_merged)
        download_manager.execute_process()
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# metadata
@relecov_tools_cli.command(help_priority=3)
@click.option(
    "-m",
    "--metadata_file",
    type=click.Path(),
    help="file containing metadata",
)
@click.option(
    "-s",
    "--sample_list_file",
    type=click.Path(),
    help="Json with the additional metadata to add to the received user metadata",
)
@click.option(
    "-o",
    "--output_dir",
    "--output-dir",
    "--output_folder",
    "--out-folder",
    "--output_location",
    "--output_path",
    "--out_dir",
    "--output",
    "output_dir",
    type=click.Path(file_okay=False, resolve_path=True),
    help="Directory where the generated output will be saved",
)
@click.option(
    "-f",
    "--files-folder",
    default=None,
    type=click.Path(),
    help="Path to folder where samples files are located",
)
@click.pass_context
def read_lab_metadata(ctx, metadata_file, sample_list_file, output_dir, files_folder):
    """
    Create the json compliant to the relecov schema from the Metadata file.
    """
    # Merge arguments
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)
    debug = ctx.obj.get("debug", False)
    new_metadata = relecov_tools.read_lab_metadata.RelecovMetadata(**args_merged)

    try:
        new_metadata.create_metadata_json()
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# validation
@relecov_tools_cli.command(help_priority=4)
@click.option("-j", "--json_file", help="Json file to validate")
@click.option(
    "-s", "--json_schema_file", help="Path to the JSON Schema file used for validation"
)
@click.option(
    "-m",
    "--metadata",
    type=click.Path(),
    help="Origin file containing metadata",
)
@click.option(
    "-o",
    "--output_dir",
    "--output-dir",
    "--output_folder",
    "--out-folder",
    "--output_location",
    "--output_path",
    "--out_dir",
    "--output",
    "output_dir",
    type=click.Path(file_okay=False, resolve_path=True),
    help="Directory where the generated output will be saved",
)
@click.option(
    "-e",
    "--excel_sheet",
    required=False,
    default=None,
    help="Optional: Name of the sheet in excel file to validate.",
)
@click.option(
    "-r",
    "--registry",
    required=False,
    default=None,
    help="Path to the JSON file containing the registered records of validated samples with their unique sample identifiers.",
)
@click.pass_context
def validate(
    ctx, json_file, json_schema_file, metadata, output_dir, excel_sheet, registry
):
    """Validate json file against schema."""
    debug = ctx.obj.get("debug", False)
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)
    validation = relecov_tools.json_validation.SchemaValidation(**args_merged)
    try:
        validation.validate()
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# send-email
@relecov_tools_cli.command(help_priority=4)
@click.option(
    "-v",
    "--validate_file",
    required=True,
    type=click.Path(exists=True),
    help="Path to the validation summary json file (validate_log_summary.json)",
)
@click.option(
    "-r",
    "--receiver_email",
    required=False,
    help="Recipient's e-mail address (optional). If not provided, it will be extracted from the institutions guide.",
)
@click.option(
    "-a",
    "--attachments",
    multiple=True,
    type=click.Path(exists=True),
    help="Path to file",
)
@click.option(
    "-t",
    "--template_path",
    type=click.Path(exists=True),
    required=False,
    default=None,
    help="Path to relecov-tools templates folder (optional)",
)
@click.option(
    "-p",
    "--email_psswd",
    help="Password for bioinformatica@isciii.es",
    required=False,
    default=None,
)
@click.option(
    "-n",
    "--additional_notes",
    type=click.Path(exists=True),
    required=False,
    help="Path to a .txt file with additional notes to include in the email (optional).",
)
@click.pass_context
def send_mail(
    ctx,
    validate_file,
    receiver_email,
    attachments,
    template_path,
    email_psswd,
    additional_notes,
):
    """
    Send a sample validation report by mail.
    """
    debug = ctx.obj.get("debug", False)
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)

    # Get arguments to use them here
    validate_file = args_merged.get("validate_file")
    receiver_email = args_merged.get("receiver_email")
    attachments = args_merged.get("attachments")
    template_path = args_merged.get("template_path")
    email_psswd = args_merged.get("email_psswd")
    additional_notes = args_merged.get("additional_notes")

    config_loader = relecov_tools.config_json.ConfigJson(extra_config=True)
    config = config_loader.get_configuration("mail_sender")
    if not config:
        raise ValueError(
            "Error: The configuration for 'mail_sender' could not be loaded."
        )

    validate_data = relecov_tools.utils.read_json_file(validate_file)
    batch = os.path.basename(os.path.dirname(os.path.abspath(validate_file)))

    if not validate_data:
        raise ValueError("Error: Validation data could not be loaded.")

    submitting_institution_code = list(validate_data.keys())[0]
    invalid_count = relecov_tools.log_summary.LogSum.get_invalid_count(validate_data)

    if not template_path:
        template_path = config.get("delivery_template_path_file")
    if not template_path or not os.path.exists(template_path):
        raise FileNotFoundError(
            "The template path could not be determined or does not exist. "
            "Please provide it via --template_path or define 'delivery_template_path_file' in the configuration."
        )

    email_sender = relecov_tools.mail.EmailSender(config, template_path)

    template_choice = click.prompt(
        "Select the type of template:\n1. Validation with errors\n2. Validation successful",
        type=int,
        default=1,
        show_choices=False,
    )
    if template_choice not in [1, 2]:
        raise ValueError("Error: invalid option.")

    # Determinar el template a usar
    if template_choice == 1:
        template_name = "template_with_errors_relecov.j2"
    else:
        template_name = "template_success_relecov.j2"

    additional_info = ""

    if additional_notes:
        with open(additional_notes, "r", encoding="utf-8") as f:
            additional_info = f.read().strip()
    else:
        if click.confirm(
            "Would you like to add a .txt file with additional notes?", default=False
        ):
            notes_path = click.prompt(
                "Enter the path to the .txt file", type=click.Path(exists=True)
            )
            with open(notes_path, "r", encoding="utf-8") as f:
                additional_info = f.read().strip()
        elif click.confirm(
            "Would you like to write additional notes manually?", default=False
        ):
            additional_info = click.prompt("Enter additional information").strip()

    institution_info = email_sender.get_institution_info(submitting_institution_code)
    if not institution_info:
        raise ValueError("Error: Could not obtain institution information.")

    institution_name = institution_info["institution_name"]
    email_receiver_from_json = institution_info["email_receiver"]

    email_body = email_sender.render_email_template(
        additional_info=additional_info,
        invalid_count=invalid_count,
        submitting_institution_code=submitting_institution_code,
        template_name=template_name,
        batch=batch,
    )

    if email_body is None:
        raise RuntimeError("Error: Could not generate mail.")

    final_receiver_email = None
    if not receiver_email:
        final_receiver_email = [
            email.strip() for email in email_receiver_from_json.split(";")
        ]
    else:
        final_receiver_email = (
            [email.strip() for email in receiver_email.split(";")]
            if isinstance(receiver_email, str)
            else receiver_email
        )

    if not final_receiver_email:
        raise ValueError("Error: Could not obtain the recipient's email address.")

    subject = (
        f"RELECOV - Informe de Validación de Muestras {batch} - {institution_name}"
    )
    try:
        email_sender.send_email(
            final_receiver_email, subject, email_body, attachments, email_psswd
        )
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# mapping to ENA schema
@relecov_tools_cli.command(help_priority=5)
@click.option("-p", "--origin_schema", help="File with the origin (relecov) schema")
@click.option(
    "-j", "--json_data", "json_file", help="File with the json data to convert"
)
@click.option(
    "-d",
    "--destination_schema",
    type=click.Choice(["ENA", "GISAID", "other"], case_sensitive=True),
    help="schema to be mapped",
)
@click.option("-f", "--schema_file", help="file with the custom schema")
@click.option(
    "-o",
    "--output_dir",
    "--output-dir",
    "--output_folder",
    "--out-folder",
    "--output_location",
    "--output_path",
    "--out_dir",
    "--output",
    "output_dir",
    type=click.Path(file_okay=False, resolve_path=True),
    help="Directory where the generated output will be saved",
)
@click.pass_context
def map(ctx, origin_schema, json_file, destination_schema, schema_file, output_dir):
    """Convert data between phage plus schema to ENA, GISAID, or any other schema"""
    debug = ctx.obj.get("debug", False)
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)
    try:
        new_schema = relecov_tools.map_schema.MappingSchema(**args_merged)
        new_schema.map_to_data_to_new_schema()
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# upload to ENA
@relecov_tools_cli.command(help_priority=6)
@click.option("-u", "--user", help="user name for login to ena")
@click.option("-p", "--password", help="password for the user to login")
@click.option("-c", "--center", help="center name")
@click.option("-e", "--ena_json", help="where the validated json is")
@click.option("-t", "--template_path", help="Path to ENA templates folder")
@click.option(
    "-a",
    "--action",
    type=click.Choice(["ADD", "MODIFY", "CANCEL", "RELEASE"], case_sensitive=False),
    help="select one of the available options",
)
@click.option("--dev", is_flag=True, default=False, help="Test submission")
@click.option("--upload_fastq", is_flag=True, default=False, help="Upload fastq files")
@click.option("-m", "--metadata_types", help="List of metadata xml types to submit")
@click.option(
    "-o",
    "--output_dir",
    "--output-dir",
    "--output_folder",
    "--out-folder",
    "--output_location",
    "--output_path",
    "--out_dir",
    "--output",
    "output_dir",
    type=click.Path(file_okay=False, resolve_path=True),
    help="Directory where the generated output will be saved",
)
@click.pass_context
def upload_to_ena(
    ctx,
    user,
    password,
    center,
    ena_json,
    template_path,
    dev,
    action,
    metadata_types,
    upload_fastq,
    output_dir,
):
    """parse data to create xml files to upload to ena"""
    debug = ctx.obj.get("debug", False)
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)
    try:
        upload_ena = relecov_tools.upload_ena_protocol.EnaUpload(**args_merged)
        upload_ena.upload()
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# upload to GISAID
@relecov_tools_cli.command(help_priority=7)
@click.option("-u", "--user", help="user name for login")
@click.option("-p", "--password", help="password for the user to login")
@click.option("-c", "--client_id", help="client-ID provided by clisupport@gisaid.org")
@click.option("-t", "--token", help="path to athentication token")
@click.option("-e", "--gisaid_json", help="path to validated json mapped to GISAID")
@click.option(
    "-i",
    "--input_path",
    help="path to fastas folder or multifasta file",
)
@click.option(
    "-o",
    "--output_dir",
    "--output-dir",
    "--output_folder",
    "--out-folder",
    "--output_location",
    "--output_path",
    "--out_dir",
    "--output",
    "output_dir",
    type=click.Path(file_okay=False, resolve_path=True),
    help="Directory where the generated output will be saved",
)
@click.option(
    "-f",
    "--frameshift",
    type=click.Choice(["catch_all", "catch_none", "catch_novel"], case_sensitive=False),
    help="frameshift notification",
)
@click.option(
    "-x",
    "--proxy_config",
    help="introduce your proxy credentials as: username:password@proxy:port",
    required=False,
)
@click.option(
    "--single",
    is_flag=True,
    default=False,
    help="input is a folder with several fasta files. Default: False",
)
@click.option(
    "--gzip",
    is_flag=True,
    default=False,
    help="input fasta is gziped. Default: False",
)
@click.pass_context
def upload_to_gisaid(
    ctx,
    user,
    password,
    client_id,
    token,
    gisaid_json,
    input_path,
    output_dir,
    frameshift,
    proxy_config,
    single,
    gzip,
):
    """parsed data to create files to upload to gisaid"""
    debug = ctx.obj.get("debug", False)
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)
    try:
        upload_gisaid = relecov_tools.gisaid_upload.GisaidUpload(**args_merged)
        upload_gisaid.gisaid_upload()
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# update_db
@relecov_tools_cli.command(help_priority=9)
@click.option("-j", "--json", help="data in json format")
@click.option(
    "-t",
    "--type",
    type=click.Choice(["sample", "bioinfodata", "variantdata"]),
    multiple=False,
    default=None,
    help="Select the type of information to upload to database",
)
@click.option(
    "-plat",
    "--platform",
    type=click.Choice(
        [
            "iskylims",
            "relecov",
        ]
    ),
    multiple=False,
    default=None,
    help="name of the platform where data is uploaded",
)
@click.option("-u", "--user", help="user name for login")
@click.option("-p", "--password", help="password for the user to login")
@click.option("-s", "--server_url", help="url of the platform server")
@click.option(
    "-f",
    "--full_update",
    is_flag=True,
    default=False,
    help="Sequentially run every update option",
)
@click.option(
    "-l",
    "--long_table",
    default=None,
    help="Long_table.json file from read-bioinfo-metadata + viralrecon",
)
@click.pass_context
def update_db(
    ctx, user, password, json, type, platform, server_url, full_update, long_table
):
    """upload the information included in json file to the database"""
    debug = ctx.obj.get("debug", False)
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)
    try:
        update_database_obj = relecov_tools.upload_database.UpdateDatabase(
            **args_merged
        )
        update_database_obj.update_db()
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# read metadata bioinformatics
@relecov_tools_cli.command(help_priority=10)
@click.option(
    "-j",
    "--json_file",
    type=click.Path(),
    help="json file containing lab metadata",
)
@click.option(
    "-s", "--json_schema_file", help="Path to the JSON Schema file used for validation"
)
@click.option("-i", "--input_folder", type=click.Path(), help="Path to input files")
@click.option(
    "-o",
    "--output_dir",
    "--output-dir",
    "--output_folder",
    "--out-folder",
    "--output_location",
    "--output_path",
    "--out_dir",
    "--output",
    "output_dir",
    type=click.Path(file_okay=False, resolve_path=True),
    help="Directory where the generated output will be saved",
)
@click.option("-s", "--software_name", help="Name of the software/pipeline used.")
@click.option(
    "--update",
    is_flag=True,
    default=False,
    help="If the output file already exists, ask if you want to update it.",
)
@click.pass_context
def read_bioinfo_metadata(
    ctx, json_file, json_schema_file, input_folder, output_dir, software_name, update
):
    """
    Create the json compliant  from the Bioinfo Metadata.
    """
    # Merge arguments
    args_merged = merge_with_extra_config(
        ctx=ctx,
        add_extra_config=True,
    )
    debug = ctx.obj.get("debug", False)
    new_bioinfo_metadata = relecov_tools.read_bioinfo_metadata.BioinfoMetadata(
        **args_merged
    )
    try:
        new_bioinfo_metadata.create_bioinfo_file()
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# metadata homogeneizer
@relecov_tools_cli.command(help_priority=12)
@click.option(
    "-i",
    "--institution",
    type=click.Choice(["isciii", "hugtip", "hunsc-iter"], case_sensitive=False),
    help="select one of the available institution options",
)
@click.option(
    "-d",
    "--directory",
    type=click.Path(),
    help="Folder where are located the additional files",
)
@click.option(
    "-o",
    "--output_dir",
    "--output-dir",
    "--output_folder",
    "--out-folder",
    "--output_location",
    "--output_path",
    "--out_dir",
    "--output",
    "output_dir",
    type=click.Path(file_okay=False, resolve_path=True),
    help="Directory where the generated output will be saved",
)
@click.pass_context
def metadata_homogeneizer(ctx, institution, directory, output_dir):
    """Parse institution metadata lab to the one used in relecov"""
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)
    debug = ctx.obj.get("debug", False)
    new_parse = relecov_tools.metadata_homogeneizer.MetadataHomogeneizer(**args_merged)
    try:
        new_parse.converting_metadata()
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# creating symbolic links
@relecov_tools_cli.command(help_priority=13)
@click.option(
    "-i",
    "--input",
    type=click.Path(),
    help="select input folder where are located the sample files",
)
@click.option(
    "-t",
    "--templates_root",
    type=click.Path(),
    help="Path to folder containing the pipeline templates from buisciii-tools",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(),
    help="select the template config file",
)
@click.option(
    "-o",
    "--output_dir",
    "--output-dir",
    "--output_folder",
    "--out-folder",
    "--output_location",
    "--output_path",
    "--out_dir",
    "--output",
    "output_dir",
    type=click.Path(file_okay=False, resolve_path=True),
    help="Directory where the generated output will be saved",
)
@click.option(
    "-f",
    "--folder_names",
    multiple=True,
    default=None,
    help="Folder basenames to process. Target folders names should match the given dates. E.g. ... -f folder1 -f folder2 -f folder3",
)
@click.pass_context
def pipeline_manager(ctx, input, templates_root, output_dir, config, folder_names):
    """
    Create the symbolic links for the samples which are validated to prepare for
    bioinformatics pipeline execution.
    """
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)
    debug = ctx.obj.get("debug", False)
    new_launch = relecov_tools.pipeline_manager.PipelineManager(**args_merged)
    try:
        new_launch.pipeline_exc()
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# schema builder
@relecov_tools_cli.command(help_priority=14)
@click.option(
    "-i",
    "--input_file",
    type=click.Path(),
    help="Path to the Excel document containing the database definition. This file must have a .xlsx extension.",
    required=True,
)
@click.option(
    "-s",
    "--schema_base",
    type=click.Path(),
    help="Path to the base schema file. This file is used as a reference to compare it with the schema generated using this module. (Default: installed schema in 'relecov-tools/relecov_tools/schema/relecov_schema.json')",
    required=False,
)
@click.option(
    "-v",
    "--draft_version",
    type=click.STRING,
    help="Version of the JSON schema specification to be used. Example: '2020-12'. See: https://json-schema.org/specification-links",
)
@click.option(
    "-d",
    "--diff",
    is_flag=True,
    help="Prints a changelog/diff between the base and incoming versions of the schema.",
)
@click.option("--version", help="Specify the schema version.")
@click.option(
    "-p", "--project", help="Specficy the project to build the metadata template."
)
@click.option(
    "--non-interactive",
    is_flag=True,
    help="Run the script without user interaction, using default values.",
)
@click.option(
    "-o",
    "--output_dir",
    "--output-dir",
    "--output_folder",
    "--out-folder",
    "--output_location",
    "--output_path",
    "--out_dir",
    "--output",
    "output_dir",
    type=click.Path(file_okay=False, resolve_path=True),
    help="Directory where the generated output will be saved",
)
@click.pass_context
def build_schema(
    ctx,
    input_file,
    schema_base,
    draft_version,
    diff,
    output_dir,
    version,
    project,
    non_interactive,
):
    """Generates and updates JSON Schema files from Excel-based database definitions."""
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)
    debug = ctx.obj.get("debug", False)
    try:
        schema_update = relecov_tools.build_schema.SchemaBuilder(**args_merged)
        new_schema = schema_update.handle_build_schema()
        if not new_schema:
            log.error("Schema build returned None. Skipping schema summary.")
            return
        schema_update.summarize_schema(new_schema)
    except Exception as e:
        if debug:
            log.exception(f"Error while building schema: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# logs to excel
@relecov_tools_cli.command(help_priority=15)
@click.option(
    "-l",
    "--lab_code",
    type=click.Path(),
    default=None,
    help="Only merge logs from target laboratory in log-summary.json files",
    required=False,
)
@click.option(
    "-o",
    "--output_dir",
    "--output-dir",
    "--output_folder",
    "--out-folder",
    "--output_location",
    "--output_path",
    "--out_dir",
    "--output",
    "output_dir",
    type=click.Path(file_okay=False, resolve_path=True),
    help="Directory where the generated output will be saved",
)
@click.option(
    "-f",
    "--files",
    help="Paths to log_summary.json files to merge into xlsx file, called once per file",
    required=True,
    multiple=True,
)
@click.pass_context
def logs_to_excel(ctx, lab_code, output_dir, files):
    """Creates a merged xlsx and Json report from all the log summary jsons given as input"""
    debug = ctx.obj.get("debug", False)
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)

    # Get arguments from merged config
    lab_code = args_merged.get("lab_code")
    output_folder = args_merged.get("output_folder")
    files = args_merged.get("files")

    all_logs = []
    full_paths = [os.path.realpath(f) for f in files]
    for file in full_paths:
        if not os.path.exists(file):
            stderr.print(f"[red]File {file} does not exist")
            log.error(f"File {file} does not exist")
            continue

        if os.path.getsize(file) == 0:
            stderr.print(f"[red]File {file} is empty")
            log.error(f"File {file} is empty")
            continue

        try:
            with open(file, "r") as f:
                content = json.load(f)
                if lab_code is not None and lab_code not in content:
                    log.warning(f"lab_code '{lab_code}' not found in {file}")
                    stderr.print(f"[yellow]lab_code '{lab_code}' not found in {file}")
                all_logs.append(content)
        except Exception as e:
            stderr.print(f"[red]Couldn't extract data from {file}: {e}")
            log.error(f"Couldn't extract data from {file}: {e}")

    if not all_logs:
        msg = f"No logs extracted. Make sure the --lab_code '{lab_code}' exists in the provided files."
        stderr.print(f"[red]{msg}")
        log.error(msg)
        raise ValueError(msg)

    logsum = relecov_tools.log_summary.LogSum(output_dir=output_dir)
    try:
        merged_logs = logsum.merge_logs(key_name=lab_code, logs_list=all_logs)
        final_logs = logsum.prepare_final_logs(logs=merged_logs)
        output_filepath = os.path.join(output_folder, lab_code + "_metadata_report")
        excel_outpath = output_filepath + ".xlsx"
        logsum.create_logs_excel(logs=final_logs, excel_outpath=excel_outpath)
        json_outpath = output_filepath + ".json"
        relecov_tools.utils.write_json_to_file(final_logs, json_outpath)
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# wrapper
@relecov_tools_cli.command(help_priority=16)
@click.option(
    "-c",
    "--config_file",
    type=click.Path(),
    help="Path to config file in yaml format [required]",
    required=True,
)
@click.option(
    "-o",
    "--output_dir",
    "--output-dir",
    "--output_folder",
    "--out-folder",
    "--output_location",
    "--output_path",
    "--out_dir",
    "--output",
    "output_dir",
    type=click.Path(file_okay=False, resolve_path=True),
    help="Directory where the generated output will be saved",
)
@click.pass_context
def wrapper(ctx, config_file, output_dir):
    """Executes the modules in config file sequentially"""
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)
    debug = ctx.obj.get("debug", False)
    process_wrapper = relecov_tools.dataprocess_wrapper.ProcessWrapper(**args_merged)
    try:
        process_wrapper.run_wrapper()
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


# upload_results
@relecov_tools_cli.command(help_priority=17)
@click.option("-u", "--user", help="User name for login to sftp server")
@click.option("-p", "--password", help="password for the user to login")
@click.option("-b", "--batch_id", help="Batch from....")
@click.option(
    "-t",
    "--template_path",
    required=False,
    help="Path to relecov-tools templates folder",
)
@click.option(
    "-r", "--project", default="Relecov", help="Project to which the samples belong"
)
@click.pass_context
def upload_results(ctx, user, password, batch_id, template_path, project):
    """Upload batch results to sftp server."""
    args_merged = merge_with_extra_config(ctx=ctx, add_extra_config=True)
    debug = ctx.obj.get("debug", False)
    upload_sftp = relecov_tools.upload_results.UploadSftp(**args_merged)
    try:
        upload_sftp.execute_process()
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


@relecov_tools_cli.command(help_priority=18)
@click.option(
    "-n",
    "--config_name",
    default=None,
    help="Name of the config key that will be added",
)
@click.option(
    "-f",
    "--config_file",
    default=None,
    help="Path to the input file: Json or Yaml format",
)
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Force replacement of existing configuration if needed",
)
@click.option(
    "--clear_config",
    is_flag=True,
    default=False,
    help="Remove given config_name from extra config: Use with empty --config_name to remove all",
)
@click.pass_context
def add_extra_config(ctx, config_name, config_file, force, clear_config):
    """Save given file content as additional configuration"""
    debug = ctx.obj.get("debug", False)
    try:
        if os.path.isfile(relecov_tools.config_json.ConfigJson._extra_config_path):
            config_json = relecov_tools.config_json.ConfigJson(extra_config=True)
        else:
            config_json = relecov_tools.config_json.ConfigJson()
        if clear_config:
            config_json.remove_extra_config(config_name)
        else:
            config_json.include_extra_config(
                config_file, config_name=config_name, force=force
            )
    except Exception as e:
        if debug:
            log.exception(f"EXCEPTION FOUND: {e}")
            raise
        else:
            log.exception(f"EXCEPTION FOUND: {e}")
            sys.exit(1)


if __name__ == "__main__":
    run_relecov_tools()



#!/usr/bin/env python
import logging
import rich

import relecov_tools.utils
from relecov_tools.read_bioinfo_metadata import BioinfoReportLog

log = logging.getLogger(__name__)
stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


def handle_pangolin_data(files_list, file_tag, pipeline_name, output_folder=None):
    """File handler to parse pangolin data (csv) into JSON structured format.

    Args:
        files_list (list): A list with paths to pangolin files.

    Returns:
        pango_data_processed: A dictionary containing pangolin data handled.
    """

    method_name = f"{handle_pangolin_data.__name__}"
    method_log_report = BioinfoReportLog()
    pango_data_processed = {}
    valid_samples = []
    try:
        files_list_processed = relecov_tools.utils.select_most_recent_files_per_sample(
            files_list
        )
        for pango_file in files_list_processed:
            try:
                pango_data = relecov_tools.utils.read_csv_file_return_dict(
                    pango_file, sep=","
                )
                pango_data_key = next(iter(pango_data))
                pango_data_updated = {
                    key.split()[0]: value for key, value in pango_data.items()
                }
                pango_data_processed.update(pango_data_updated)
                valid_samples.append(pango_data_key.split()[0])
            except (FileNotFoundError, IndexError) as e:
                method_log_report.update_log_report(
                    method_name,
                    "warning",
                    f"Error occurred while processing file {pango_file}: {e}",
                )
                continue
    except Exception as e:
        method_log_report.update_log_report(
            method_name, "warning", f"Error occurred while processing files: {e}"
        )
    if len(valid_samples) > 0:
        method_log_report.update_log_report(
            method_name,
            "valid",
            f"Successfully handled data in samples: {', '.join(valid_samples)}",
        )
    method_log_report.print_log_report(method_name, ["valid", "warning"])
    return pango_data_processed


def quality_control_evaluation(data):
    """Evaluates QC status for each sample based on predefined thresholds.

    Parameters:
    -----------
    data : list of dict
        List of sample metadata dictionaries containing metrics such as coverage,
        ambiguity, number of Ns, %LDMutations, etc.

    Returns:
    --------
    list of dict
        The same list with an added 'qc_test' field per sample:
        - 'pass' if all evaluable conditions are met
        - 'fail' if any condition fails
        - Adds 'qc_failed' field with reasons for failure
    """
    log_report = BioinfoReportLog()
    method_name = f"{quality_control_evaluation.__name__}"

    thresholds = {
        "per_sgene_ambiguous": ("<", 10.0),
        "per_sgene_coverage": (">", 98.0),
        "per_ldmutations": (">", 60.0),  # except if 'Not Evaluable'
        "number_of_sgene_frameshifts": ("==", 0),
        "number_of_unambiguous_bases": (">", 24000),
        "number_of_Ns": ("<", 5000),
        "pass_reads": (">", 50000),
        "per_reads_host": ("<", 20.0),
    }

    def is_not_evaluable(value):
        return isinstance(value, str) and "Not Evaluable" in value

    def invert_operator(op):
        return {">": "<", "<": ">", ">=": "<=", "<=": ">=", "==": "!=", "!=": "=="}.get(
            op, f"NOT_{op}"
        )

    conditions = {
        k: (
            (
                lambda x, op=op, th=th: (
                    True
                    if is_not_evaluable(x) and k == "per_ldmutations"
                    else (
                        eval(f"{float(x)} {op} {th}")
                        if isinstance(x, (int, float))
                        else False
                    )
                )
            )
        )
        for k, (op, th) in thresholds.items()
    }

    data, warnings = relecov_tools.assets.pipeline_utils.utils.evaluate_qc_samples(
        data,
        thresholds,
        conditions,
        invert_operator,
        is_not_evaluable,
    )

    for warn in warnings:
        log_report.update_log_report(method_name, warn)

    return data



import logging

# from pyparsing import col
import rich.console
import sys
import pandas as pd
import os

import relecov_tools.utils
from Bio import SeqIO
from relecov_tools.config_json import ConfigJson


# import site


log = logging.getLogger(__name__)
stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class GisaidUpload:
    def __init__(
        self,
        user=None,
        password=None,
        client_id=None,
        token=None,
        gisaid_json=None,
        input_path=None,
        output_dir=None,
        frameshift=None,
        proxy_config=None,
        single=False,
        gzip=False,
    ):
        if (
            token is None
        ):  # borrar comentario: solo si no existe el token necesita user, passwd y client_id
            self.token = None
            print("Token is not introduced, creating a new one...")
            if user is None:
                self.user = relecov_tools.utils.prompt_text(
                    msg="Enter your username defined in GISAID"
                )
            else:
                self.user = user
            if password is None:
                self.passwd = relecov_tools.utils.prompt_password(
                    msg="Enter your password to GISAID"
                )
            else:
                self.passwd = password
            if client_id is None:
                self.client_id = relecov_tools.utils.prompt_password(
                    msg="Enter your client-ID to GISAID. Email clisupport@gisaid.org to request client-ID"
                )
            else:
                self.client_id = client_id
        else:
            self.token = token
        if gisaid_json is None:
            self.gisaid_json = relecov_tools.utils.prompt_path(
                msg="Select the GISAID json file to upload"
            )
        else:
            self.gisaid_json = gisaid_json
        if output_dir is None:
            self.output_dir = relecov_tools.utils.prompt_path(
                msg="Select the folder to store the log files"
            )
        else:
            self.output_dir = output_dir
        if input_path is None:
            self.fasta_path = relecov_tools.utils.prompt_path(
                msg="Select path to fasta file/s"
            )
        else:
            self.fasta_path = input_path
        if frameshift is None:
            self.frameshift = relecov_tools.utils.prompt_selection(
                msg="Select frameshift notification",
                choices=["catch_all", "catch_novel", "catch_none"],
            )
        else:
            self.frameshift = frameshift
        # Add proxy settings: username:password@proxy:port (optional)
        if proxy_config is None:
            # borrar comentario: este mensaje no me convence
            self.proxy_config = None
            print("Proxy configuration is not set")
        else:
            self.proxy_config = proxy_config
        self.single = single
        self.gzip = gzip

    # Metadatos

    def complete_mand_fields(self, dataframe):
        """Complete mandatory empty fields with 'unknown'"""
        dataframe.loc[dataframe["covv_gender"] == "", "covv_gender"] = "unknown"
        dataframe.loc[dataframe["covv_patient_age"] == "", "covv_patient_age"] = (
            "unknown"
        )

        authors = [authors_field for authors_field in dataframe["covv_authors"]]
        if "" in authors or "unknown" in authors:
            log.error("Invalid value for author. This field is required in full")
            stderr.print(
                "[red] Invalid value for authors. This field is required in full, 'unknown' is not allowed"
            )
            sys.exit(1)

        dataframe.loc[dataframe["covv_subm_lab_addr"] == "", "covv_subm_lab_addr"] = (
            "unknown"
        )
        dataframe.loc[dataframe["covv_subm_lab"] == "", "covv_subm_lab"] = "unknown"
        dataframe.loc[dataframe["covv_orig_lab_addr"] == "", "covv_orig_lab_addr"] = (
            "unknown"
        )
        dataframe.loc[dataframe["covv_orig_lab"] == "", "covv_orig_lab"] = "unknown"
        dataframe.loc[dataframe["covv_patient_status"] == "", "covv_patient_status"] = (
            "unknown"
        )
        dataframe.loc[dataframe["covv_type"] == "", "covv_type"] = "betacoronavirus"
        dataframe.loc[dataframe["covv_passage"] == "", "covv_passage"] = "Original"

        config_json = ConfigJson()
        gisaid_config = config_json.get_configuration("GISAID_configuration")
        submitter_id = gisaid_config["submitter"]
        dataframe.loc[dataframe["submitter"] == "", "submitter"] = submitter_id

        bioinfo_config = config_json.get_configuration("bioinfo_analysis")
        assembly_method = bioinfo_config["fixed_values"][
            "bioinformatics_protocol_software_name"
        ]
        dataframe.loc[
            dataframe["covv_assembly_method"] == "", "covv_assembly_method"
        ] = assembly_method

        return dataframe

    def metadata_to_csv(self):
        """Transform metadata json to csv"""
        data = relecov_tools.utils.read_json_file(self.gisaid_json)
        df_data = pd.DataFrame(data)

        config_json = ConfigJson()
        fields = config_json.get_configuration("gisaid_csv_headers")
        col_df = list(df_data.columns)
        for field in fields:
            if field not in col_df:
                df_data.insert(4, field, "")

        config_lab_json = ConfigJson()
        lab_json_conf = config_lab_json.get_topic_data(
            "lab_metadata", "laboratory_data"
        )
        lab_json_file = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "conf", lab_json_conf["file"]
        )
        lab_json = relecov_tools.utils.read_json_file(lab_json_file)
        for lab in lab_json:
            for i in range(len(df_data)):
                if lab["collecting_institution"] == df_data["covv_orig_lab"][i]:
                    df_data["covv_location"][i] = " / ".join(
                        [
                            "Europe",
                            lab["geo_loc_country"],
                            lab["geo_loc_state"],
                            lab["geo_loc_city"],
                        ]
                    )

        df_data.replace("not provided", "unknown", inplace=True)
        df_data_comp = self.complete_mand_fields(df_data)
        df_data_path = os.path.join(self.output_dir, "meta_gisaid.csv")
        if not os.path.exists(self.output_dir):
            os.mkdir(self.output_dir)
        df_data_comp.to_csv(df_data_path, index=False)
        return df_data_path

    # generar template con cli3
    # ADD TOKEN WARNING and file token  .authtoken
    # add bash from cli3
    """
    os.system(
        "cli3 upload --database EpiCoV --token ./gisaid.authtoken --metadata gisaid_template.csv  --fasta multi.fasta --frameshift (OPTIONAL, default: catch_all) --failed --proxy --log"
    )
    cli3 upload
    --database EpiCoV
    --token ./gisaid.authtoken
    --metadata gisaid_template.csv
    --fasta multi.fasta
    --frameshift (OPTIONAL, default: catch_all)
    --failed default creates file failed.out where the failed records will be
    --proxy
    --log default creates file failed.out where the log will be )

    """

    def create_multifasta(self):
        """Create multifasta from single fastas (if --single)"""
        if self.single:
            gather_fastas_path = os.path.join(self.fasta_path, "*.fa*")
            if self.gzip:
                os.system(
                    "zcat %s > %s/multifasta.fasta"
                    % (gather_fastas_path, self.output_dir)
                )
            else:
                os.system(
                    "cat %s > %s/multifasta.fasta"
                    % (gather_fastas_path, self.output_dir)
                )
            multifasta = "%s/multifasta.fasta" % self.output_dir

        else:
            if self.gzip:
                os.system(
                    "zcat %s > %s/multifasta.fasta" % (self.fasta_path, self.output_dir)
                )
                multifasta = "%s/multifasta.fasta" % self.output_dir
            else:
                multifasta = self.fasta_path
        return multifasta

    def change_headers(self, multifasta):
        """Transform multifasta ids/headers to GISAID format"""
        data = relecov_tools.utils.read_json_file(self.gisaid_json)
        virus_name = [name["covv_virus_name"] for name in data]
        multi_gis_path = os.path.join(
            self.output_dir, "processed_multifasta_gisaid.fasta"
        )
        with open(multifasta) as old_fasta, open(multi_gis_path, "a") as new_fasta:
            records = SeqIO.parse(old_fasta, "fasta")
            for record in records:
                for name in virus_name:
                    if name.split("/")[-2].split("-")[-1] in record.id:
                        record.id = name
                        record.description = name
                        SeqIO.write(record, new_fasta, "fasta")
        return multi_gis_path

    def cli3_auth(self):
        """Create authenticate token"""
        os.system(
            "cli3 authenticate --username %s --password %s --client_id %s"
            % (self.user, self.passwd, self.client_id)
        )
        self.token = "gisaid.authtoken"

    def cli3_upload(self):
        """Upload to GISAID"""
        if self.proxy_config is None:
            os.system(
                "cli3 upload --token %s --metadata %s --fasta %s --frameshift %s"
                % (
                    self.token,
                    self.metadata_to_csv(),
                    self.change_headers(self.create_multifasta()),
                    self.frameshift,
                )
            )
        else:
            os.system(
                "cli3 upload --token %s --metadata %s --fasta %s --frameshift %s --proxy %s"
                % (
                    self.token,
                    self.metadata_to_csv(),
                    self.change_headers(self.create_multifasta()),
                    self.frameshift,
                    self.proxy_config,
                )
            )

    def gisaid_upload(self):
        """Upload to GISAID"""
        if self.token is None:
            self.cli3_auth()
        self.cli3_upload()

    """"
    Upload
    Subir con cli3
    Token
    Opción de configurar proxy
    def upload(self):
    Create the required files and upload to ENA
    self.convert_input_json_to_ena()
    self.create_structure_to_ena()
    """



#!/usr/bin/env python
import os
import sys
import logging
import rich.console

import relecov_tools.utils
from relecov_tools.config_json import ConfigJson

log = logging.getLogger(__name__)
stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class MetadataHomogeneizer:
    """MetadataHomogeneizer object"""

    def __init__(self, institution=None, directory=None, output_dir=None):
        # open config
        self.config_json = ConfigJson()
        # read heading from config
        self.heading = self.config_json.get_topic_data(
            "lab_metadata", "metadata_lab_heading"
        )

        # handle institution
        if institution is None:
            self.institution = relecov_tools.utils.prompt_selection(
                msg="Select the available mapping institution",
                choices=["isciii", "hugtip", "hunsc-iter"],
            ).upper()
        else:
            self.institution = institution.upper()

        mapping_json_file = os.path.join(
            os.path.dirname(__file__),
            "schema",
            "institution_schemas",
            self.config_json.get_topic_data(
                "institution_mapping_file", self.institution
            ),
        )

        self.mapping_json_data = relecov_tools.utils.read_json_file(mapping_json_file)

        if directory is None:
            directory = relecov_tools.utils.prompt_path(
                msg="Select the directory which contains additional files for metadata"
            )
        if not os.path.exists(directory):
            log.error("Folder for additional files %s does not exist ", directory)
            stderr.print(
                "[red] Folder for additional files " + directory + " does not exist"
            )
            sys.exit(1)

        try:
            lab_metadata = self.mapping_json_data["required_files"]["metadata_file"][
                "file_name"
            ]
        except KeyError:
            log.error("Metadata File is not defined in schema")
            stderr.print("[red] Metadata File is not defined in schema")
            sys.exit(1)

        metadata_path = os.path.join(directory, lab_metadata)

        if not os.path.isfile(metadata_path):
            log.error("Metadata File %s does not exists", metadata_path)
            stderr.print("[red] Metadata File " + metadata_path + "does not exists")
            sys.exit(1)
        self.lab_metadata = self.mapping_json_data["required_files"]["metadata_file"]
        self.lab_metadata["file_name"] = metadata_path

        self.additional_files = []

        if len(self.mapping_json_data["required_files"]) > 1:
            for key, values in self.mapping_json_data["required_files"].items():
                if key == "metadata_file":
                    continue
                if values["file_name"] == "":
                    self.additional_files.append(values)
                    continue
                f_path = os.path.join(directory, values["file_name"])
                if not os.path.isfile(f_path):
                    log.error("Additional file %s does not exist ", f_path)
                    stderr.print("[red] Additional file " + f_path + " does not exist")
                    sys.exit(1)
                values["file_name"] = f_path
                self.additional_files.append(values)

        # Check if python file is defined
        function_file = self.mapping_json_data["python_file"]

        if function_file == "":
            self.function_file = None
        else:
            self.function_file = os.path.join(
                os.path.dirname(__file__), "institution_scripts", function_file
            )
            if not os.path.isfile(self.function_file):
                log.error("File with functions %s does not exist ", self.function_file)
                stderr.print(
                    "[red] File with functions "
                    + self.function_file
                    + " does not exist"
                )
                sys.exit(1)
        if output_dir is None:
            self.output_dir = relecov_tools.utils.prompt_path(
                msg="Select the output folder"
            )
        else:
            self.output_dir = output_dir
        self.processed_metadata = False

    def mapping_metadata(self, ws_data):
        map_fields = self.mapping_json_data["required_files"]["metadata_file"][
            "mapped_fields"
        ]
        map_data = []
        for row in ws_data:
            row_data = {}
            for dest_map, orig_map in map_fields.items():
                row_data[dest_map] = row[orig_map]
            map_data.append(row_data)

        return map_data

    def add_fixed_fields(self, mapped_data):
        add_data = [self.heading]
        fixed_fields = self.mapping_json_data["fixed_fields"]
        for row in mapped_data:
            new_row_data = []
            for field in self.heading:
                if field in row:
                    data = row[field]
                elif field in fixed_fields:
                    data = fixed_fields[field]
                else:
                    data = ""
                new_row_data.append(data)
            add_data.append(new_row_data)
        return add_data

    def handling_files(self, file_data, data_to_add):
        """Added information based on the required file configuration.
        The first time this function is called is for mapping the laboratory
        metadata to ISCIII. For this time mapping_metadata method is used.
        and return the list that is going to be used later for adding/modifing
        information
        """
        if file_data["file_name"] != "":
            f_name = file_data["file_name"]
            stderr.print("[blue] Starting processing file " + f_name)
            if f_name.endswith(".json"):
                data = relecov_tools.utils.read_json_file(f_name)
            elif f_name.endswith(".tsv"):
                data = relecov_tools.utils.read_csv_file_return_dict(f_name, "\t")
            elif f_name.endswith(".csv"):
                data = relecov_tools.utils.read_csv_file_return_dict(f_name, ",")
            elif f_name.endswith(".xlsx"):
                header_flag = self.metadata_processing.get("header_flag")
                data = relecov_tools.utils.read_excel_file(
                    f_name, "Sheet", header_flag, leave_empty=True
                )
            else:
                log.error("Additional file extension %s is not supported ", f_name)
                stderr.print(
                    "[red] Additional file extension " + f_name + " is not supported"
                )
                sys.exit(1)
        else:
            data = ""
        if not self.processed_metadata:
            self.processed_metadata = True
            return self.mapping_metadata(data)

        if file_data["function"] == "None":
            mapping_idx = self.heading.index(file_data["mapped_key"])
            for row in data_to_add[1:]:
                # new_row_data = []
                s_value = str(row[mapping_idx])
                try:
                    item_data = data[s_value]
                except KeyError:
                    log.info(
                        "Additional file %s does not have the information for %s ",
                        f_name,
                        s_value,
                    )
                    stderr.print(
                        "[yellow] Additional file "
                        + f_name
                        + " does not have information for "
                        + str(s_value)
                    )
                    continue
                    # sys.exit(1)
                for m_field, f_field in file_data["mapped_fields"].items():
                    try:
                        meta_idx = self.heading.index(m_field)
                    except ValueError as e:
                        log.error("Field %s does not exist in Metadata ", e)
                        stderr.print(f"[red] Field {e} does not exist")
                        sys.exit(1)
                    row[meta_idx] = item_data[f_field]

        else:
            func_name = file_data["function"]
            stderr.print("[yellow] Start processing function " + func_name)
            exec(
                "from relecov_tools.institution_scripts."
                + self.institution
                + " import "
                + func_name
            )
            # somehow this overrides additional_data working as a pointer
            eval(
                func_name
                + "(data_to_add, data, file_data['mapped_fields'], self.heading)"
            )

        stderr.print("[green] Succesful processing of additional file ")
        return data_to_add

    def converting_metadata(self):
        stderr.print("[blue] Reading the metadata file to convert")

        # metadata_file contains the primary source of information. First we map it.
        mapped_data = self.handling_files(self.lab_metadata, "")
        stderr.print("[green] Successful conversion mapping to ISCIII metadata")
        stderr.print("[blue] Adding fixed information")

        # Then we add the fixed data
        additional_data = self.add_fixed_fields(mapped_data)

        # Fetch the additional files and include the information in metadata
        stderr.print("[blue] reading and mapping de information that are in files")
        for additional_file in self.additional_files:
            additional_data = self.handling_files(additional_file, additional_data)

        # write to excel mapped data
        f_name = os.path.join(self.output_dir, "converted_metadata_lab.xlsx")
        stderr.print("[blue] Dumping information to excel")
        post_process = {"insert_rows": 3, "insert_cols": 1}
        relecov_tools.utils.write_to_excel_file(
            additional_data, f_name, "METADATA_LAB", post_process
        )
        stderr.print("[green] Complete process for mapping to ISCIII metadata")
        return



#!/usr/bin/env python
import copy
import json
import os
import yaml
import warnings
import rich.console
import paramiko
import pandas as pd
import relecov_tools.utils
import relecov_tools.sftp_client
from datetime import datetime
from itertools import islice
from secrets import token_hex
from csv import writer as csv_writer, Error as CsvError
from openpyxl import load_workbook as openpyxl_load_workbook
from pandas import read_excel, ExcelWriter, concat
from pandas.errors import ParserError, EmptyDataError
from relecov_tools.config_json import ConfigJson
from relecov_tools.base_module import BaseModule

# from relecov_tools.rest_api import RestApi

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class MetadataError(Exception):
    def __init__(self, message):
        super().__init__(message)


class DownloadManager(BaseModule):
    def __init__(
        self,
        user=None,
        password=None,
        conf_file=None,
        download_option=None,
        output_dir=None,
        target_folders=None,
        subfolder=None,
    ):
        """Initializes the sftp object"""
        super().__init__(output_dir=output_dir, called_module="download")
        self.log.info("Initiating download process")
        config_json = ConfigJson(extra_config=True)
        self.allowed_file_ext = config_json.get_topic_data(
            "sftp_handle", "allowed_file_extensions"
        )
        sftp_user = user
        sftp_passwd = password
        self.target_folders = target_folders
        self.subfolder = subfolder
        self.allowed_download_options = config_json.get_topic_data(
            "sftp_handle", "allowed_download_options"
        )
        if download_option not in self.allowed_download_options:
            self.download_option = relecov_tools.utils.prompt_selection(
                "Options", self.allowed_download_options
            )
        else:
            self.download_option = download_option
        if conf_file is None:
            # self.sftp_server = config_json.get_topic_data("sftp_handle", "sftp_server")
            # self.sftp_port = config_json.get_topic_data("sftp_handle", "sftp_port")
            self.platform_storage_folder = config_json.get_topic_data(
                "sftp_handle", "platform_storage_folder"
            )
            self.abort_if_md5_mismatch = (
                True
                if config_json.get_topic_data("sftp_handle", "abort_if_md5_mismatch")
                == "True"
                else False
            )
        else:
            if not os.path.isfile(conf_file):
                self.log.error("Configuration file %s does not exists", conf_file)
                stderr.print(
                    "[red] Configuration file does not exist. " + conf_file + "!"
                )
                raise FileNotFoundError(f"Config file {conf_file} does not exists")
            with open(conf_file, "r") as fh:
                config = yaml.load(fh, Loader=yaml.FullLoader)
            try:
                # self.sftp_server = config["sftp_server"]
                # self.sftp_port = config["sftp_port"]
                self.target_folders = config["target_folders"]
                try:
                    self.platform_storage_folder = config["platform_storage_folder"]
                except KeyError:
                    self.platform_storage_folder = config_json.get_topic_data(
                        "sftp_handle", "platform_storage_folder"
                    )
                sftp_user = config["sftp_user"]
                sftp_passwd = config["sftp_passwd"]
            except KeyError as e:
                self.log.error("Invalid configuration file. Missing %s", e)
                stderr.print(f"[red] Invalid configuration file. Missing {e} !")
                raise ValueError(f"Invalid configuration file. Missing {e}")
        if output_dir is not None:
            if os.path.isdir(output_dir):
                self.platform_storage_folder = os.path.realpath(output_dir)
            else:
                self.log.error("Output location does not exist, aborting")
                stderr.print("[red] Output location does not exist, aborting")
                raise FileNotFoundError(f"Output dir does not exist {output_dir}")
        if sftp_user is None:
            sftp_user = relecov_tools.utils.prompt_text(msg="Enter the user id")
        if isinstance(self.target_folders, str):
            self.target_folders = self.target_folders.split(",")
        self.logsum = self.parent_log_summary(output_dir=self.platform_storage_folder)
        if sftp_passwd is None:
            sftp_passwd = relecov_tools.utils.prompt_password(msg="Enter your password")
        self.metadata_lab_heading = config_json.get_topic_data(
            "lab_metadata", "metadata_lab_heading"
        )
        self.metadata_processing = config_json.get_topic_data(
            "sftp_handle", "metadata_processing"
        )
        self.avoidable_characters = config_json.get_topic_data(
            "sftp_handle", "skip_when_found"
        )
        self.samples_json_fields = config_json.get_topic_data(
            "lab_metadata", "samples_json_fields"
        )
        # initialize the sftp client
        self.relecov_sftp = relecov_tools.sftp_client.SftpRelecov(
            conf_file, sftp_user, sftp_passwd
        )
        self.finished_folders = {}
        self.set_batch_id(datetime.today().strftime("%Y%m%d%H%M%S"))
        self.defer_cleanup = False

    def create_local_folder(self, folder):
        """Create folder to download files in local path using date

        Args:
            folder (str): name of remote folder to be downloaded

        Returns:
            local_folder_path(str): path to the new created folder
        """
        self.log.info("Creating folder %s to download files", folder)
        platform_storage_folder = self.platform_storage_folder
        path_parts = folder.split("/")
        cod_folder = path_parts[0]
        batch_folder = path_parts[-1].replace("_tmp_processing", "")
        local_folder_path = os.path.join(
            platform_storage_folder, cod_folder, batch_folder
        )
        os.makedirs(local_folder_path, exist_ok=True)
        self.log.info("Created the folder to download files %s", local_folder_path)
        return local_folder_path

    def get_remote_folder_files(self, folder, local_folder, file_list):
        """Create the subfolder with the present date and fetch all files from
        the remote sftp server

        Args:
            folder (str): name of remote folder to be downloaded
            local_folder (str): name of local folder to store downloaded files
            file_list (list(str)): list of files in remote folder to be downloaded

        Returns:
            fetched_files(list(str)): list of successfully downloaded files
        """

        fetched_files = list()
        self.log.info("Trying to fetch files in remote server")
        stderr.print(f"Fetching {len(file_list)} files from {folder}")
        for file in file_list:
            file_to_fetch = os.path.join(folder, os.path.basename(file))
            output_file = os.path.join(local_folder, os.path.basename(file))
            if self.relecov_sftp.get_from_sftp(
                file_to_fetch, output_file, exist_ok=True
            ):
                fetched_files.append(os.path.basename(file))
            else:
                # Try to download again n times
                for _ in range(3):
                    if self.relecov_sftp.get_from_sftp(file_to_fetch, output_file):
                        fetched_files.append(os.path.basename(file))
                        break
                else:
                    self.log.warning(
                        "Couldn't fetch %s from %s after 3 tries", file, folder
                    )
        return fetched_files

    def find_remote_md5sum(self, folder, pattern="md5sum"):
        """Search for a pattern in remote folder, by default is md5sum

        Args:
            folder (str): folder path in remote repository
            pattern (str, optional): Regex used to find file. Defaults to "md5sum".

        Returns:
            md5_file(str): file basename if found. If not found returns False
        """
        all_files = self.relecov_sftp.get_file_list(folder)
        md5_file = [file for file in all_files if pattern in file]
        if len(md5_file) == 1:
            return md5_file[0]
        else:
            return False

    def verify_md5_checksum(self, local_folder, fetched_files, fetched_md5):
        """Check if the md5 value from sftp matches with the one generated locally"""
        required_retransmition = []
        successful_files = []
        # fetch the md5 file if exists
        self.log.info("Searching for local md5 file")
        stderr.print("[blue]Verifying file integrity in md5 hashes")
        avoid_chars = self.avoidable_characters
        hash_dict = relecov_tools.utils.read_md5_checksum(fetched_md5, avoid_chars)
        if not hash_dict:
            error_text = "md5sum file could not be read, md5 hashes won't be validated"
            self.include_warning(error_text)
            return fetched_files, False
        # check md5 checksum for each file
        for f_name in hash_dict.keys():
            if f_name not in fetched_files:
                # Skip those files in md5sum that were not downloaded by any reason
                continue
            f_path = os.path.join(local_folder, f_name)
            if hash_dict[f_name] == relecov_tools.utils.calculate_md5(f_path):
                successful_files.append(f_name)
                self.log.info("Successful file download for %s", f_name)
            else:
                required_retransmition.append(f_name)
                self.log.warning("%s requested file re-sending", f_name)
        return successful_files, required_retransmition

    def create_files_with_metadata_info(
        self, local_folder, samples_dict, md5_dict, metadata_file
    ):
        """Copy metadata file from folder, extend samples_dict with md5hash for
        each file. Then create a Json file with this dict

        Args:
            local_folder (str): Path to folder with downloaded files and output
            samples_dict (dict{str:str}): same structure as validate_remote_files()
            md5_dict (dict(str:str)): Zipped dict of files_list and md5hash_list
            metadata_file (str): Name of the downloaded metadata file to rename it
        """
        samples_to_delete = []
        lab_code = local_folder.split("/")[-2]
        # TODO: Move these prefixes to configuration.json
        new_meta_file = self.tag_filename("lab_metadata_" + lab_code + ".xlsx")
        sample_data_file = self.tag_filename("samples_data_" + lab_code + ".json")
        sample_data_path = os.path.join(local_folder, sample_data_file)
        os.rename(metadata_file, os.path.join(local_folder, new_meta_file))
        error_text = "Sample %s incomplete. Not added to final Json"

        data = copy.deepcopy(samples_dict)
        for sample, values in data.items():
            if not all(val for val in values):
                self.include_error(str(error_text % sample), sample)
                samples_to_delete.append(sample)
                continue
            # TODO: Move these keys to configuration.json
            values["sequence_file_path_R1"] = local_folder
            values["sequence_file_R1_md5"] = md5_dict.get(values["sequence_file_R1"])
            if values.get("sequence_file_R2"):
                values["sequence_file_path_R2"] = local_folder
                values["sequence_file_R2_md5"] = md5_dict.get(
                    values["sequence_file_R2"]
                )
            values["batch_id"] = self.batch_id
        if samples_to_delete:
            data = {k: v for k, v in data.items() if k not in samples_to_delete}
        with open(sample_data_path, "w", encoding="utf-8") as fh:
            fh.write(json.dumps(data, indent=4, sort_keys=True, ensure_ascii=False))
        # Feed accessible dict with necessary information for wrapper to work
        self.log.info("Successfully created samples json file %s", sample_data_path)
        return

    def remove_duplicated_values(self, sample_file_dict):
        """remove keys that share the same value due to duplication in sample_dict

        Args:
            sample_file_dict (dict(str:dict(str:str))): dictionary with sample_name
            as keys and a dict for both R1 filename and/or R2 if paired-end reads
            and fastq-file paths. e.g. {sample1:{sequence_file_path_R1:sample1.fastq.gz}}

        Returns:
            clean_sample_dict: sample_dictionary without duplications in values
        """
        inverted_dict = {}
        for sample, fastq_dict in sample_file_dict.items():
            # Dictionary values are not hashable, so you need to create a tuple of them
            samp_fastqs = tuple(fastq_dict.values())
            # Setting values as keys to find those samples refering to the same file
            for fastq in samp_fastqs:
                inverted_dict[fastq] = inverted_dict.get(fastq, []) + [sample]
        duplicated_dict = {k: v for k, v in inverted_dict.items() if len(v) > 1}
        dup_samples_list = [samp for dups in duplicated_dict.values() for samp in dups]
        non_duplicated_keys = {
            k: v for k, v in sample_file_dict.items() if k not in dup_samples_list
        }
        clean_sample_dict = {key: sample_file_dict[key] for key in non_duplicated_keys}
        clean_duplicated_dict = {
            k: [samp for samp in v if "_remove_" not in samp]
            for k, v in inverted_dict.items()
            if len(v) > 1
        }
        unique_dup_samples_list = list(
            {samp for dups in clean_duplicated_dict.values() for samp in dups}
        )
        if dup_samples_list:
            error_text = "Multiple samples in metadata pointing to the same file: %s"
            self.include_warning(error_text % clean_duplicated_dict)
            stderr.print(f"[yellow]{error_text % clean_duplicated_dict}")
            stderr.print(
                "[yellow]These samples won't be processed: ", unique_dup_samples_list
            )
            for fastq, samples in clean_duplicated_dict.items():
                [self.include_error(str(error_text % fastq), samp) for samp in samples]

        return clean_sample_dict

    def read_metadata_file(self, meta_f_path, return_data=True):
        """Read Excel file, check if the header matches with the one defined in config.

        Args:
            meta_f_path (str): Path to the Excel file.

        Raises:
            MetadataError: If the header in the Excel is different from config.

        Returns:
            ws_metadata_lab: worksheet (openpyxl or simulated from pandas)
            metadata_header: column names of the header
            header_row: row where the header is located in the sheet (1-based)
        """

        warnings.simplefilter(action="ignore", category=UserWarning)
        header_flag = self.metadata_processing.get("header_flag")
        sheet_name = self.metadata_processing.get("excel_sheet")
        meta_column_list = self.metadata_lab_heading

        try:
            wb_file = openpyxl_load_workbook(meta_f_path, data_only=True)
            ws_metadata_lab = wb_file[sheet_name]
            try:
                header_row = [
                    i + 1
                    for i, x in enumerate(ws_metadata_lab.values)
                    if header_flag in x
                ][0]
            except IndexError:
                error_text = "Header could not be found for excel file %s"
                raise MetadataError(str(error_text % os.path.basename(meta_f_path)))
            for cell in ws_metadata_lab[header_row]:
                if cell.value is not None:
                    cell.value = cell.value.strip()
            metadata_header = [
                x.value for x in ws_metadata_lab[header_row] if x.value is not None
            ]
            if meta_column_list != metadata_header[1:]:
                diffs = [
                    x
                    for x in set(metadata_header[1:] + meta_column_list)
                    if x not in meta_column_list or x not in metadata_header
                ]
                self.log.error(
                    "Config field metadata_lab_heading is different from .xlsx header"
                )
                stderr.print(
                    "[red]Header in metadata file is different from config file, aborting"
                )
                stderr.print("[red]Differences: ", diffs)
                raise MetadataError(f"Metadata header different from config: {diffs}")
            if return_data:
                return ws_metadata_lab, metadata_header, header_row
            else:
                return True

        except Exception as openpyxl_error:
            self.log.warning(
                f"openpyxl failed to read the Excel file: {openpyxl_error}"
            )
            self.log.warning("Attempting to read using pandas fallback")

            try:
                df = pd.read_excel(meta_f_path, sheet_name=sheet_name, header=None)
                header_row_mask = df.apply(
                    lambda row: header_flag in row.values, axis=1
                )
                if not header_row_mask.any():
                    raise MetadataError(
                        f"Header flag '{header_flag}' not found in {meta_f_path}"
                    )
                header_row = header_row_mask.idxmax()
                metadata_header = [
                    str(h).strip() for h in df.iloc[header_row] if pd.notna(h)
                ]

                if meta_column_list != metadata_header[1:]:
                    diffs = [
                        x
                        for x in set(metadata_header[1:] + meta_column_list)
                        if x not in meta_column_list or x not in metadata_header
                    ]
                    self.log.error(
                        "Config field metadata_lab_heading is different from .xlsx header"
                    )
                    stderr.print(
                        "[red]Header in metadata file is different from config file, aborting"
                    )
                    stderr.print("[red]Differences: ", diffs)
                    raise MetadataError(
                        f"Metadata header different from config: {diffs}"
                    )

                if return_data:
                    ws_metadata_lab = df.iloc[header_row + 1 :].values.tolist()
                    return (
                        ws_metadata_lab,
                        metadata_header,
                        header_row + 1,
                    )  # +1 to be consistent with openpyxl
                else:
                    return True

            except Exception as pandas_error:
                raise MetadataError(
                    f"Failed to read metadata Excel file with both openpyxl and pandas:\n"
                    f"- openpyxl error: {openpyxl_error}\n"
                    f"- pandas error: {pandas_error}"
                )

    def get_sample_fastq_file_names(self, local_folder, meta_f_path):
        """Read excel metadata template and create dictionary with files for each sample

        Args:
            local_folder (str): folder where the excel file has been downloaded
            meta_f_path (str): path to the downloaded excel file with metadata

        Returns:
            clean_sample_dict(dict(str:{str:str})): Nested dictionary for each sample
            {sample1: {"sequence_file_R1": "sample1_R1.fastq.gz",
                        "sequence_file_R2": "sample1_R2.fastq.gz"},
             sample2:{...} }
        """

        def set_nones_to_str(row, req_vals):
            row = list(row)
            for index in req_vals:
                if row[index] is None:
                    row[index] = ""
            return row

        if not os.path.isfile(meta_f_path):
            self.log.error("Metadata file does not exist on %s", local_folder)
            stderr.print("[red] METADATA_LAB.xlsx do not exist in" + local_folder)
            return False
        sample_file_dict = {}
        metadata_ws, meta_header, header_row = self.read_metadata_file(meta_f_path)
        # TODO Include these columns in config
        index_sampleID = meta_header.index("Sample ID given for sequencing")
        index_altID = meta_header.index("Sample ID given by originating laboratory")
        index_layout = meta_header.index("Library Layout")
        index_fastq_r1 = meta_header.index("Sequence file R1")
        index_fastq_r2 = meta_header.index("Sequence file R2")
        req_vals = [index_layout, index_fastq_r1, index_fastq_r2]
        counter = header_row
        for row in islice(metadata_ws.values, header_row, metadata_ws.max_row):
            row = set_nones_to_str(row, req_vals)
            counter += 1
            if not row[index_sampleID]:
                if not row[index_altID]:
                    sample_id = row[index_altID]
                else:
                    sample_id = row[index_fastq_r1].split(".")[0]
            else:
                sample_id = row[index_sampleID]
            if sample_id:
                try:
                    s_name = str(sample_id).strip()
                except ValueError as e:
                    self.log.error("Unable to convert to string. %s", e)
                    stderr.print("[red]Unable to convert to string. ", e)
                    continue
                if s_name in sample_file_dict:
                    if row[index_fastq_r1] == sample_file_dict[s_name].get(
                        "sequence_file_R1", ""
                    ) or row[index_fastq_r2] == sample_file_dict[s_name].get(
                        "sequence_file_R2", ""
                    ):
                        s_name = s_name + "_remove_" + str(counter)
                    else:
                        log_text = "Found more samples with the same Sample ID given for sequencing. Only the first one remains."
                        stderr.print(log_text)
                        self.include_warning(log_text, sample=s_name)
                        continue
                if not row[index_fastq_r1]:
                    log_text = "Sequence File R1 not defined in Metadata for sample %s"
                    stderr.print(f"[red]{str(log_text % s_name)}")
                    self.include_error(entry=str(log_text % s_name), sample=s_name)
                try:
                    if (
                        "paired" in row[index_layout].lower()
                        and not row[index_fastq_r2]
                    ):
                        error_text = "Sample %s is paired-end, but no R2 given"
                        self.include_error(error_text % str(sample_id), s_name)
                    if "single" in row[index_layout].lower() and row[index_fastq_r2]:
                        error_text = "Sample %s is single-end, but R1 and R2 were given"
                        self.include_error(error_text % str(sample_id), s_name)
                except AttributeError:
                    error_text = (
                        f"Missing or invalid 'Library Layout' value for sample {s_name}. "
                        f"Please check the metadata file and provide a value (single/paired)."
                    )
                    self.include_error(error_text, sample=s_name)
                    stderr.print(f"[red]{error_text}")
                    raise MetadataError(error_text)

                sample_file_dict[s_name] = {}
                # TODO: move these keys to configuration.json
                sample_file_dict[s_name]["sequence_file_R1"] = row[
                    index_fastq_r1
                ].strip()
                if row[index_fastq_r2]:
                    sample_file_dict[s_name]["sequence_file_R2"] = row[
                        index_fastq_r2
                    ].strip()
            else:
                txt = f"Sample for row {counter} in metadata skipped. No sample ID nor file provided"
                self.include_warning(entry=txt)
        # Remove duplicated files
        clean_sample_dict = self.remove_duplicated_values(sample_file_dict)
        return clean_sample_dict

    def get_metadata_file(self, remote_folder, local_folder):
        """Check if the metadata file exists

        Args:
            remote_folder (str): path to the folder in remote repository
            local_folder (str): path to the local folder

        Raises:
            FileNotFoundError: If missing metadata excel file or merging error.

        Returns:
            local_meta_file: Path to downloaded metadata file / merged metadata file.
        """
        remote_files_list = self.relecov_sftp.get_file_list(remote_folder)
        meta_files = [fi for fi in remote_files_list if fi.endswith(".xlsx")]

        def download_remote_metafile(target_meta_file):
            local_meta_file = os.path.join(
                local_folder, os.path.basename(target_meta_file)
            )
            try:
                self.relecov_sftp.get_from_sftp(target_meta_file, local_meta_file)
            except (IOError, PermissionError) as e:
                raise type(e)(f"[red]Unable to fetch metadata file {e}")
            self.log.info(
                "Obtained metadata file %s from %s",
                local_meta_file,
                remote_folder,
            )
            return local_meta_file

        if not meta_files:
            raise FileNotFoundError(f"Missing metadata file for {remote_folder}")
        os.makedirs(self.platform_storage_folder, exist_ok=True)
        if len(meta_files) > 1:
            # Merging multiple excel files into a single one
            self.log.warning(
                f"[yellow]Merging multiple metadata files in {remote_folder}"
            )
            metadata_ws = self.metadata_processing.get("excel_sheet")
            header_flag = self.metadata_processing.get("header_flag")
            local_meta_list = []
            for remote_metafile in meta_files:
                local_meta_file = download_remote_metafile(remote_metafile)
                local_meta_list.append(local_meta_file)
            meta_df_list = []
            for loc_meta in local_meta_list:
                try:
                    loc_meta_df = self.excel_to_df(loc_meta, metadata_ws, header_flag)
                    meta_df_list.append(loc_meta_df)
                except (ParserError, EmptyDataError, MetadataError, KeyError) as e:
                    error_text = f"Could not process {os.path.basename(loc_meta)}: {e}"
                    self.include_error(error_text)
                os.remove(loc_meta)
            if meta_df_list:
                merged_df = meta_df_list[0]
            else:
                raise MetadataError("No single metadata file could be merged")
            for meta_df in meta_df_list[1:]:
                merged_df = self.merge_metadata(metadata_ws, merged_df, meta_df)
            folder_name = os.path.dirname(local_meta_file)
            excel_name = str(folder_name.split("/")[-1]) + "merged_metadata.xlsx"
            merged_excel_path = os.path.join(folder_name, excel_name)
            pd_writer = ExcelWriter(merged_excel_path, engine="xlsxwriter")
            for sheet in merged_df.keys():
                format_sheet = merged_df[sheet].astype(str)
                format_sheet.replace("nan", None, inplace=True)
                format_sheet.to_excel(pd_writer, sheet_name=sheet, index=False)
            pd_writer.close()
            local_meta_file = merged_excel_path
            return merged_excel_path
        else:
            target_meta_file = meta_files[0]
            local_meta_file = download_remote_metafile(target_meta_file)
        return local_meta_file

    def validate_remote_files(self, remote_folder, local_folder):
        """Check if the files in the remote folder are the ones defined in metadata file

        Args:
            remote_folder (str): Name of remote folder being validated
            local_folder (str): Name of folder where files are being downloaded

        Raises:
            FileNotFoundError: If none of the files in remote folder are valid

        Returns:
            sample_files_dict (dict): same structure as self.get_sample_fastq_file_names
            local_meta_file (str): location of downloaded metadata excel file
        """
        local_meta_file = self.get_metadata_file(remote_folder, local_folder)
        out_folder = os.path.dirname(local_meta_file)
        allowed_extensions = self.allowed_file_ext
        remote_files_list = [
            os.path.basename(file)
            for file in self.relecov_sftp.get_file_list(remote_folder)
        ]
        filtered_files_list = sorted(
            [fi for fi in remote_files_list if fi.endswith(tuple(allowed_extensions))]
        )
        sample_files_dict = self.get_sample_fastq_file_names(
            out_folder, local_meta_file
        )
        # Include the samples in the process log summary
        for sample in sample_files_dict.keys():
            self.include_new_key(sample=sample)
        metafiles_list = sorted(
            sum([list(fi.values()) for _, fi in sample_files_dict.items()], [])
        )
        if sorted(filtered_files_list) == sorted(metafiles_list):
            self.log.info("Files in %s match with metadata file", remote_folder)
        else:
            log_text = "Some files in %s do not match the ones described in metadata"
            self.log.warning(log_text % remote_folder)
            stderr.print(f"[gold1]{log_text % remote_folder}")
            set_list = set(metafiles_list)
            mismatch_files = [fi for fi in filtered_files_list if fi not in set_list]
            mismatch_rev = [fi for fi in set_list if fi not in filtered_files_list]

            if mismatch_files:
                error_text1 = "Files in folder missing in metadata: %s"
                self.include_warning(error_text1 % str(mismatch_files))
            if mismatch_rev:
                error_text2 = "Files in metadata missing in folder: %s"
                self.include_warning(error_text2 % str(mismatch_rev))
            # Try to check if the metadata filename lacks the proper extension
            self.log.info("Trying to match files without proper file extension")
            sample_files_dict = self.process_filedict(
                sample_files_dict, filtered_files_list
            )
        if not any(value for value in sample_files_dict.values()):
            raise FileNotFoundError(
                "No files from metadata found in %s" % remote_folder
            )
        self.log.info("Finished validating files based on metadata")
        stderr.print("[blue]Finished validating files based on metadata")
        return sample_files_dict, local_meta_file

    def delete_remote_files(self, remote_folder, files=None, skip_seqs=False):
        """Delete files from remote folder.

        Args:
            remote_folder (str): Path to folder in remote repository.
            files (list(str), optional): List of target filenames in remote repository.
            skip_seqs (bool, optional): Skip sequencing files based on extension.
        """
        self.log.info(f"Deleting files in remote {remote_folder}.")
        stderr.print(f"[blue]Deleting files in remote {remote_folder}...")
        all_files = self.relecov_sftp.get_file_list(remote_folder)
        file_paths = {os.path.basename(f): f for f in all_files}
        if files is None:
            files_to_remove = all_files
        else:
            files_to_remove = files
        for file in files_to_remove:
            if skip_seqs and file.endswith(tuple(self.allowed_file_ext)):
                continue
            matched_path = file_paths.get(os.path.basename(file))
            if matched_path:
                try:
                    self.relecov_sftp.remove_file(matched_path)
                except (IOError, PermissionError) as e:
                    self.log.error(f"Could not delete remote file {matched_path}: {e}")
                    stderr.print(
                        f"[red]Could not delete remote file {matched_path}. Error: {e}"
                    )
            else:
                self.log.warning(f"File not found before deletion: {file}")
                stderr.print(f"[red]File not found before deletion: {file}")
        return

    def rename_remote_folder(self, remote_folder):
        if "tmp_processing" in remote_folder:
            new_name = remote_folder.replace("tmp_processing", "invalid_samples")
            if new_name == remote_folder:
                self.log.warning("Remote folder %s was already renamed", remote_folder)
                return
            try:
                self.relecov_sftp.rename_file(remote_folder, new_name)
                if self.finished_folders.get(remote_folder):
                    self.finished_folders[new_name] = self.finished_folders.pop(
                        remote_folder
                    )
                self.log.info(
                    "Successfully renamed %s to %s" % (remote_folder, new_name)
                )
                self.logsum.rename_log_key(remote_folder, new_name)
            except (OSError, PermissionError) as e:
                log_text = f"Could not rename remote {remote_folder}. Error: {e}"
                self.log.error(log_text)
        else:
            self.log.warning(
                "No `tmp_processing` pattern in %s, not renamed" % remote_folder
            )
        return

    def clean_remote_folder(self, remote_folder):
        """Delete a folder from remote sftp, check if it is empty or not first.

        Args:
            remote_folder (str): path to folder in remote repository
        """

        def remove_client_dir(remote_folder, top_level):
            # Never remove a folder in the top level
            if len(remote_folder.replace("./", "").split("/")) >= top_level:
                self.log.info("Trying to remove %s", remote_folder)
                try:
                    self.relecov_sftp.remove_dir(remote_folder)
                    self.log.info("Successfully removed %s", remote_folder)
                except (OSError, PermissionError) as e:
                    log_text = f"Could not delete remote {remote_folder}. Error: {e}"
                    self.log.error(log_text)
                    stderr.print(log_text)
            else:
                self.log.info("%s is a top-level folder. Not removed", remote_folder)

        remote_folder_files = self.relecov_sftp.get_file_list(remote_folder)
        if remote_folder_files:
            self.rename_remote_folder(remote_folder)
            log_text = f"Remote folder {remote_folder} not empty. Not removed."
            self.log.warning(log_text)
        else:
            if self.subfolder:
                top_level = 3
            else:
                top_level = 2
            remove_client_dir(remote_folder, top_level)
        return

    def move_processing_fastqs(self, folders_with_metadata):
        """Gather all the files from any subfolder into a processing folder

        Args:
            folders_with_metadata (dict(str:list)): Dictionary updated from merge_md5sums()

        Returns:
            folders_with_metadata (dict(str:list)): Same dict updated with files successfully moved
        """
        self.log.info("Moving remote files to each temporal processing folder")
        stderr.print("[blue]Moving remote files to each temporal processing folder")
        for folder, files in folders_with_metadata.items():
            self.current_folder = folder.split("/")[0]
            successful_files = []
            unique_files = list(set(files))
            for file in unique_files:
                if not file.endswith(tuple(self.allowed_file_ext)):
                    continue

                file_dest = os.path.join(folder, os.path.basename(file))
                if file_dest in successful_files:
                    continue
                try:
                    self.relecov_sftp.rename_file(file, file_dest)
                    successful_files.append(file_dest)
                except OSError as e:
                    self.log.error(f"Error moving file {file} to {file_dest}: {e}")
                    stderr.print(f"[red]Error moving file {file} to {file_dest}: {e}")
            folders_with_metadata[folder] = successful_files
        return folders_with_metadata

    def merge_md5sums(self, folders_with_metadata):
        """Download the md5sums for each folder, merge them into a single one,
        upload them to the remote processing folder.

        Args:
            folders_with_metadata (dict(str:list)): Dictionary with remote folders
            and their files. All subfolder filenames are merged into a single key.

        Raises:
            FileNotFoundError: If no md5sum file is found in the folder

        Returns:
            folders_with_metadata: Same dict updated with the merged md5sum file
        """
        output_dir = self.platform_storage_folder

        # TODO: Include this function in relecov_tools.utils
        def md5_merger(md5_filelist, avoid_chars=None):
            """Merge all md5 files from a given list into a single multi-line md5sum"""
            md5dict_list = []
            for md5sum in md5_filelist:
                hash_dict = relecov_tools.utils.read_md5_checksum(md5sum, avoid_chars)
                if hash_dict:
                    md5dict_list.append(hash_dict)
            # Sort hashes and files back to the original order.
            merged_md5 = {
                md5: file for mdict in md5dict_list for file, md5 in mdict.items()
            }
            return merged_md5

        def md5_handler(md5sumlist, output_dir):
            """Download all the remote md5sum files in a list, merge them
            into a single md5checksum and upload it back to sftp"""
            downloaded_md5files = []
            for md5sum in md5sumlist:
                md5_name = "_".join([token_hex(nbytes=12), "md5_temp.md5"])
                fetched_md5 = os.path.join(output_dir, md5_name)
                if self.relecov_sftp.get_from_sftp(
                    file=md5sum, destination=fetched_md5
                ):
                    downloaded_md5files.append(fetched_md5)
            merged_md5 = md5_merger(downloaded_md5files, self.avoidable_characters)
            if merged_md5:
                merged_name = "_".join([folder.split("/")[0], "md5sum.md5"])
                merged_md5_path = os.path.join(output_dir, merged_name)
                with open(merged_md5_path, "w") as md5out:
                    write_md5 = csv_writer(md5out, delimiter="\t")
                    write_md5.writerows(merged_md5.items())
                md5_dest = os.path.join(folder, os.path.basename(merged_md5_path))
                self.relecov_sftp.upload_file(merged_md5_path, md5_dest)
                # Remove local files once merged and uploaded
                os.remove(merged_md5_path)
                [os.remove(md5_file) for md5_file in downloaded_md5files]
                return md5_dest
            else:
                error_text = "No md5sum could be processed in remote folder"
                raise FileNotFoundError(error_text)

        for folder, files in folders_with_metadata.items():
            self.current_folder = folder.split("/")[0]
            self.log.info("Merging md5sum files from %s...", self.current_folder)
            stderr.print(f"[blue]Merging md5sum files from {self.current_folder}...")
            md5flags = [".md5", "md5sum", "md5checksum"]
            md5sumlist = [fi for fi in files if any(flag in fi for flag in md5flags)]
            if not md5sumlist:
                error_text = "No md5sum could be found in remote folder %s"
                stderr.print(f"[yellow]{error_text % folder}")
                self.include_warning(error_text % folder)
                continue
            folders_with_metadata[folder] = [fi for fi in files if fi not in md5sumlist]
            try:
                uploaded_md5 = md5_handler(md5sumlist, output_dir)
            except (FileNotFoundError, OSError, PermissionError, CsvError) as e:
                error_text = "Could not merge md5files for %s. Reason: %s"
                stderr.print(f"[yellow]{error_text % (self.current_folder, str(e))}")
                self.include_warning(error_text % (self.current_folder, str(e)))
                continue
            if uploaded_md5:
                folders_with_metadata[folder].append(uploaded_md5)

        return folders_with_metadata

    def merge_metadata(self, meta_sheet=None, *metadata_tables):
        """Merge a variable number of metadata dataframes to the first one. Merge them
        only into a certain sheet from a multi-sheet excel file if sheetname is given.

        Args:
            meta_sheet (str): Name of the sheet containing metadata in excel file
            *metadata_tables (list(pandas.DataFrame)): Dataframes to be merged

        Returns:
            merged_df (pandas.DataFrame): A merged dataframe from the given tables
        """
        for idx, table in enumerate(metadata_tables):
            if idx == 0:
                merged_df = table
                continue
            if meta_sheet:
                merged_df[meta_sheet] = concat(
                    [merged_df[meta_sheet], table[meta_sheet]], ignore_index=True
                ).drop_duplicates()
            else:
                merged_df = concat(
                    [merged_df, table], ignore_index=True
                ).drop_duplicates()
        return merged_df

    def excel_to_df(self, excel_file, metadata_sheet, header_flag):
        """Read an excel file, return a dict with a dataframe for each sheet in it.
        Process the given sheet with metadata, removing all rows until header is found

        Args:
            excel_file (str): Path to the local excel file with metadata
            metadata_sheet (str): Name of the sheet containing metadata in excel file
            header_flag (str): Name of one of the columns from the metadata header

        Raises:
            MetadataError: If no header could be found matching header flag

        Returns:
            excel_df (dict(str:pandas.DataFrame)): Dict {name_of_excel_sheet:DataFrame}
            containing all sheets in the excel file as pandas dataframes.
        """
        # Get every sheet from the first excel file
        excel_df = read_excel(excel_file, dtype=str, sheet_name=None)
        meta_df = excel_df[metadata_sheet]
        if header_flag in meta_df.columns:
            return excel_df
        header_row = None
        for idx in range(len(meta_df)):
            if any(meta_df.loc[idx, x] == header_flag for x in meta_df.columns):
                header_row = idx
        meta_df.columns = meta_df.iloc[header_row]
        excel_df[metadata_sheet] = meta_df.drop(meta_df.index[: (header_row + 1)])
        excel_df[metadata_sheet] = excel_df[metadata_sheet].reset_index(drop=True)
        return excel_df

    def merge_subfolders(self, target_folders):
        """For each first-level folder in the sftp, merge all the subfolders within
        it in a single one called '*_tmp_processing' by moving all the fastq files from
        them. Merge the metadata excel and md5 files from each subfolder too.

        Args:
            target_folders (dict(str:list)): Dictionary with folders and their files

        Returns:
            clean_target_folders (dict(str:list)): Dict with '*_tmp_processing' folders
            and their content. All subfolder filenames are merged into a single key.
        """
        metadata_ws = self.metadata_processing.get("excel_sheet")
        header_flag = self.metadata_processing.get("header_flag")
        output_dir = self.platform_storage_folder
        date_and_time = self.batch_id
        exts = self.allowed_file_ext

        def upload_merged_df(merged_excel_path, last_main_folder, merged_df):
            """Upload metadata dataframe merged from all subfolders back to sftp"""
            self.relecov_sftp.make_dir(last_main_folder)
            pd_writer = ExcelWriter(merged_excel_path, engine="xlsxwriter")
            for sheet in merged_df.keys():
                format_sheet = merged_df[sheet].astype(str)
                format_sheet.replace("nan", None, inplace=True)
                format_sheet.to_excel(pd_writer, sheet_name=sheet, index=False)
            pd_writer.close()
            dest = os.path.join(last_main_folder, os.path.basename(merged_excel_path))
            self.relecov_sftp.upload_file(merged_excel_path, dest)
            os.remove(merged_excel_path)
            return

        def pre_validate_folder(folder, folder_files):
            """Check if remote folder has sequencing files and a valid metadata file"""
            if not any(file.endswith(tuple(exts)) for file in folder_files):
                error_text = "Remote folder %s skipped. No sequencing files found."
                self.include_error(error_text % folder)
                return
            try:
                downloaded_metadata = self.get_metadata_file(folder, output_dir)
            except (FileNotFoundError, OSError, PermissionError, MetadataError) as err:
                error_text = "Remote folder %s skipped. Reason: %s"
                self.include_error(error_text % (folder, err))
                return
            try:
                self.read_metadata_file(downloaded_metadata, return_data=False)
            except (MetadataError, KeyError) as excel_error:
                error_text = f"Folder {self.current_folder} skipped: %s"
                os.remove(downloaded_metadata)
                self.include_error(error_text % excel_error)
                return
            return downloaded_metadata

        folders_with_metadata = {}
        processed_folders = []
        merged_df = merged_excel_path = last_main_folder = excel_name = None
        self.log.info("Setting %s remote folders...", str(len(target_folders.keys())))
        stderr.print(f"[blue]Setting {len(target_folders.keys())} remote folders...")
        for folder in sorted(target_folders.keys()):
            if "invalid_samples" in folder:
                self.log.warning("Skipped invalid_samples folder %s", folder)
                continue
            self.current_folder = folder
            # Include the folder in the final process log summary
            downloaded_metadata = pre_validate_folder(folder, target_folders[folder])
            if not downloaded_metadata:
                continue
            # Create a temporal name to avoid duplicated filenames
            meta_filename = "_".join([folder.split("/")[-1], "metadata_temp.xlsx"])
            local_meta = os.path.join(output_dir, meta_filename)
            os.rename(downloaded_metadata, local_meta)

            # Taking the main folder for each lab as reference for merge and logs
            main_folder = folder.split("/")[0]
            self.current_folder = main_folder
            if self.subfolder is not None:
                tmp_folder_parent = os.path.join(main_folder, self.subfolder)
            else:
                tmp_folder_parent = main_folder
            temporal_foldername = f"{date_and_time}_tmp_processing"
            temp_folder = os.path.join(tmp_folder_parent, temporal_foldername)
            # Get every file except the excel ones as they are going to be merged
            filelist = [fi for fi in target_folders[folder] if not fi.endswith(".xlsx")]
            if not folders_with_metadata.get(temp_folder):
                log_text = "Trying to merge metadata from %s in %s"
                self.log.info(log_text % (main_folder, temp_folder))
                stderr.print(f"[blue]{log_text % (main_folder, temp_folder)}")
                if merged_df:
                    # Write the previous merged metadata df before overriding it
                    try:
                        upload_merged_df(merged_excel_path, last_main_folder, merged_df)
                        folders_with_metadata[last_main_folder].append(excel_name)
                    except OSError:
                        error_text = "Error uploading merged metadata back to sftp: %s"
                        self.include_error(error_text % last_main_folder)
                        del folders_with_metadata[last_main_folder]
                try:
                    merged_df = self.excel_to_df(local_meta, metadata_ws, header_flag)
                except (ParserError, EmptyDataError, MetadataError, KeyError) as e:
                    meta_name = os.path.basename(downloaded_metadata)
                    error_text = "%s skipped. Error while processing excel %s: %s"
                    self.include_error(error_text % (main_folder, meta_name, str(e)))
                    os.remove(local_meta)
                    continue
                folders_with_metadata[temp_folder] = []
                folders_with_metadata[temp_folder].extend(filelist)
                # rename metadata file to avoid filename duplications
                excel_name = "_".join([folder.split("/")[0], "merged_metadata.xlsx"])
                merged_excel_path = os.path.join(output_dir, excel_name)
                os.rename(local_meta, merged_excel_path)
                # Keep a track of the main_folder for next iteration
                last_main_folder = temp_folder
            else:
                # If temp_folder has subfolders in it, merge everything
                folders_with_metadata[temp_folder].extend(filelist)
                new_df = self.excel_to_df(local_meta, metadata_ws, header_flag)
                merged_df = self.merge_metadata(metadata_ws, merged_df, new_df)
                os.remove(local_meta)
            processed_folders.append(folder)
        # End of loop

        # Write last dataframe to file once loop is finished
        if folders_with_metadata.get(last_main_folder):
            if excel_name not in folders_with_metadata[last_main_folder]:
                upload_merged_df(merged_excel_path, last_main_folder, merged_df)
                folders_with_metadata[last_main_folder].append(excel_name)

        # Merge md5files and upload them to tmp_processing folder
        merged_md5_folders = self.merge_md5sums(folders_with_metadata)
        # Move all the files from each subfolder into its tmp_processing folder
        clean_target_folders = self.move_processing_fastqs(merged_md5_folders)
        log_text = "Remote folders merged into %s folders. Proceed with processing"
        self.log.info(log_text % len(clean_target_folders.keys()))
        stderr.print(f"[green]{log_text % len(clean_target_folders.keys())}")
        return clean_target_folders, processed_folders

    def select_target_folders(self):
        """Find the selected folders in remote if given, else select every folder.

        Returns:
            folders_to_process (dict(str:list)): Dictionary with folders and their files
        """
        root_directory_list = self.relecov_sftp.list_remote_folders(".", recursive=True)
        clean_root_list = [folder.replace("./", "") for folder in root_directory_list]
        if not root_directory_list:
            self.log.error("Error while listing folders in remote. Aborting")
            raise ConnectionError("Error while listing folders in remote")
        if self.target_folders is None:
            target_folders = clean_root_list
        elif self.target_folders[0] == "ALL":
            self.log.info("Showing folders from remote SFTP for user selection")
            target_folders = relecov_tools.utils.prompt_checkbox(
                msg="Select the folders that will be targeted",
                choices=sorted(clean_root_list),
            )
        else:
            target_folders = [tf for tf in self.target_folders if tf in clean_root_list]
        if self.subfolder is not None:
            new_target_folders = []
            for tfolder in target_folders:
                if self.subfolder in self.relecov_sftp.list_remote_folders(tfolder):
                    # Lets dive directly into the selected subfolder
                    new_target_folders.append(os.path.join(tfolder, self.subfolder))
                else:
                    self.log.debug(
                        f"Folder {tfolder} did not have subfolder {self.subfolder}. Skipped"
                    )
            target_folders = new_target_folders.copy()
        if not target_folders:
            self.log.error(
                "No remote folders matching selection %s", self.target_folders
            )
            stderr.print("Found no remote folders matching selection")
            stderr.print(f"List of remote folders: {str(clean_root_list)}")
            raise ValueError(
                f"Found no remote folders matching selection {self.target_folders}"
            )
        folders_to_process = {}
        for targeted_folder in target_folders:
            try:
                subfolders = self.relecov_sftp.list_remote_folders(
                    targeted_folder, recursive=True
                )
            except (FileNotFoundError, OSError) as e:
                self.log.error(
                    f"Error during sftp listing. {targeted_folder} skipped:", e
                )
                continue
            for folder in subfolders:
                list_files = self.relecov_sftp.get_file_list(folder)
                if list_files:
                    folders_to_process[folder] = list_files
                else:
                    self.log.error(f"Subfolder {folder} empty. Skipped")
        if len(folders_to_process) == 0:
            self.log.info("Exiting process, folders were empty.")
            self.log.error("There are no files in the selected folders.")
            self.relecov_sftp.close_connection()
            raise FileNotFoundError(f"Target folders are empty: {target_folders}")
        return folders_to_process

    def compress_and_update(self, fetched_files, files_to_compress, local_folder):
        """compress the given list of files_to_compress and update files_list

        Args:
            fetched_files (list(str)): list of all downloaded files
            files_to_compress (list(str)): list of files that are uncompressed

        Returns:
            fetched_files(list(str)): files list including the new compressed files
        """
        compressed_files = list()
        for file in files_to_compress:
            f_path = os.path.join(local_folder, file)
            compressed = relecov_tools.utils.compress_file(f_path)
            if not compressed:
                error_text = "Could not compress file %s, file not found" % str(file)
                self.include_error(error_text, f_path)
                continue
            # Remove file after compression is completed
            compressed_files.append(file)
            try:
                os.remove(f_path)
            except (FileNotFoundError, PermissionError) as e:
                self.log.warning(f"Could not delete file: {e}")
        fetched_files = [
            (fi + ".gz" if fi in compressed_files else fi) for fi in fetched_files
        ]
        return fetched_files

    def process_filedict(
        self, valid_filedict, clean_fetchlist, corrupted=[], md5miss=[]
    ):
        """Process the dictionary from validate_remote_files() to update filenames
        and remove samples that failed any validation process.

        Args:
            valid_filedict (dict{str:str}): same structure as validate_remote_files()
            clean_fetchlist (list(str)): List of files that passed validation process
            processed (bool): Indicates if filedict has been processed previously

        Returns:
            processed(dict{str:str}): Updated valid_filedict
        """
        processed_dict = {}
        error_text = "corrupted or md5 mismatch for %s"
        warning_text = "File %s not found in md5sum. Creating hash"
        for sample, vals in valid_filedict.items():
            processed_dict[sample] = {}
            for key, val in vals.items():
                processed_dict[sample][key] = None
                if val in corrupted:
                    self.include_error(error_text % val, sample=sample)
                if val in md5miss:
                    self.include_warning(warning_text % val, sample=sample)
                for file in clean_fetchlist:
                    if val in file:
                        processed_dict[sample][key] = file
            # remove sample if it has missing files
            if not all(x in clean_fetchlist for x in processed_dict[sample].values()):
                if not corrupted:
                    error_text = "Sample %s skipped: missing files in sftp"
                    self.include_error(str(error_text % sample), sample=sample)
                del processed_dict[sample]
        return processed_dict

    def download(self, target_folders):
        """Manages all the different functions to download files, verify their
        integrity and create initial json with filepaths and md5 hashes

        Args:
            target_folders (dict): dictionary
        """
        main_folder = self.platform_storage_folder
        try:
            os.makedirs(main_folder, exist_ok=True)
        except OSError as e:
            self.log.error("You do not have permissions to create folder %s", e)
            raise
        folders_to_download = target_folders
        for folder in folders_to_download.keys():
            self.current_folder = folder.split("/")[0]
            # Close previously open connection to avoid timeouts
            try:
                self.relecov_sftp.close_connection()
            except paramiko.SSHException:
                pass
            # Check if the connection has been closed due to time limit
            self.relecov_sftp.open_connection()
            self.log.info("Processing folder %s", folder)
            stderr.print("[blue]Processing folder " + folder)
            # Validate that the files are the ones described in metadata.

            local_folder = self.create_local_folder(folder)
            try:
                valid_filedict, meta_file = self.validate_remote_files(
                    folder, local_folder
                )
            except (FileNotFoundError, IOError, PermissionError, MetadataError) as fail:
                self.log.error("%s, skipped", fail)
                stderr.print(f"[red]{fail}, skipped")
                self.include_error(fail)
                continue
            # Get the files in each folder
            files_to_download = [
                fi for vals in valid_filedict.values() for fi in vals.values()
            ]
            fetched_files = self.get_remote_folder_files(
                folder, local_folder, files_to_download
            )
            if not fetched_files:
                error_text = "No files could be downloaded in folder %s" % str(folder)
                stderr.print(f"{error_text}")
                self.include_error(error_text)
                continue
            self.log.info("Finished download for folder: %s", folder)
            stderr.print(f"Finished download for folder {folder}")
            remote_md5sum = self.find_remote_md5sum(folder)
            if remote_md5sum:
                # Get the md5checksum to validate integrity of files after download
                fetched_md5 = os.path.join(
                    local_folder, os.path.basename(remote_md5sum)
                )
                self.relecov_sftp.get_from_sftp(
                    file=remote_md5sum, destination=fetched_md5
                )
                successful_files, corrupted = self.verify_md5_checksum(
                    local_folder, fetched_files, fetched_md5
                )
                # try to download the files again to discard errors during download
                if corrupted:
                    self.log.info("Found md5 mismatches, downloading again.")
                    stderr.print("[gold1]Found md5 mismatches, downloading again...")
                    self.get_remote_folder_files(folder, local_folder, corrupted)
                    saved_files, corrupted = self.verify_md5_checksum(
                        local_folder, corrupted, fetched_md5
                    )
                    if saved_files:
                        successful_files.extend(saved_files)
                    if corrupted:
                        error_text = "Found corrupted files: %s. Removed"
                        stderr.print(f"[red]{error_text % (str(corrupted))}")
                        self.include_warning(error_text % (str(corrupted)))
                        if self.abort_if_md5_mismatch:
                            error_text = "Stop processing %s due to corrupted files."
                            stderr.print(f"[red]{error_text % folder}")
                            self.include_error(error_text % "folder")
                            relecov_tools.utils.delete_local_folder(local_folder)
                            continue
                hash_dict = relecov_tools.utils.read_md5_checksum(
                    fetched_md5, self.avoidable_characters
                )
                self.log.info("Finished md5 check for folder: %s", folder)
                stderr.print(f"[blue]Finished md5 verification for folder {folder}")
            else:
                corrupted = []
                error_text = "No single md5sum file could be found in %s" % folder
                stderr.print(f"[red]{error_text}")
                self.include_warning(error_text)

            to_remove = set()
            for sample_id, files in list(valid_filedict.items()):
                if any(
                    files.get(key) in corrupted
                    for key in ["sequence_file_R1", "sequence_file_R2"]
                ):
                    to_remove.update(files.values())

            # Delete corrupted files before proceeding
            for file_name in to_remove:
                path = os.path.join(local_folder, file_name)
                try:
                    os.remove(path)
                    self.log.info(
                        "File %s was removed because it was corrupted", file_name
                    )
                    corrupted.append(file_name)
                except (FileNotFoundError, PermissionError, OSError) as e:
                    error_text = "Could not remove corrupted file %s: %s"
                    self.log.error(error_text % (path, e))
                    stderr.print(f"[red]{error_text % (path, e)}")

            seqs_fetchlist = [
                fi for fi in fetched_files if fi.endswith(tuple(self.allowed_file_ext))
            ]
            seqs_fetchlist = [fi for fi in seqs_fetchlist if fi not in corrupted]
            # Checking for uncompressed files
            files_to_compress = [
                fi
                for fi in seqs_fetchlist
                if not fi.endswith(".gz") and not fi.endswith(".bam")
            ]
            if files_to_compress:
                comp_files = str(len(files_to_compress))
                self.log.info("Found %s uncompressed files, compressing...", comp_files)
                stderr.print(f"Found {comp_files} uncompressed files, compressing...")
                clean_fetchlist = self.compress_and_update(
                    seqs_fetchlist, files_to_compress, local_folder
                )
            else:
                clean_fetchlist = seqs_fetchlist
            clean_pathlist = [os.path.join(local_folder, fi) for fi in clean_fetchlist]

            for file in clean_fetchlist:
                full_f_path = os.path.join(local_folder, file)
                if not relecov_tools.utils.check_gzip_integrity(full_f_path):
                    corrupted.append(file)

            not_md5sum = []
            if remote_md5sum:
                # Get hashes from provided md5sum, create them for those not provided
                files_md5_dict = {}
                for path in clean_pathlist:
                    f_name = os.path.basename(path)
                    if f_name in corrupted:
                        clean_fetchlist.remove(f_name)
                    elif f_name in successful_files:
                        files_md5_dict[f_name] = hash_dict[f_name]
                    else:
                        if not str(f_name).rstrip(".gz") in files_to_compress:
                            error_text = "File %s not found in md5sum. Creating hash"
                            self.log.warning(error_text % f_name)
                            not_md5sum.append(f_name)
                        else:
                            self.log.info(
                                "File %s was compressed, creating md5hash", f_name
                            )
                        files_md5_dict[f_name] = relecov_tools.utils.calculate_md5(path)
            else:
                md5_hashes = [
                    relecov_tools.utils.calculate_md5(path) for path in clean_pathlist
                ]
                files_md5_dict = dict(zip(clean_fetchlist, md5_hashes))
            files_md5_dict = {
                x: y for x, y in files_md5_dict.items() if x not in corrupted
            }
            processed_filedict = self.process_filedict(
                valid_filedict, clean_fetchlist, corrupted=corrupted, md5miss=not_md5sum
            )
            self.create_files_with_metadata_info(
                local_folder, processed_filedict, files_md5_dict, meta_file
            )
            if self.logsum.logs.get(self.current_folder):
                self.logsum.logs[self.current_folder].update({"path": local_folder})
                try:
                    log_name = (
                        self.tag_filename("download_" + self.current_folder)
                        + "_log_summary.json"
                    )
                    self.parent_create_error_summary(
                        filepath=os.path.join(local_folder, log_name),
                        logs={
                            self.current_folder: self.logsum.logs[self.current_folder]
                        },
                    )
                except Exception as e:
                    self.log.error(
                        "Could not create logsum for %s: %s" % (folder, str(e))
                    )
            self.log.info(f"Finished processing {folder}")
            stderr.print(f"[green]Finished processing {folder}")
            self.finished_folders[folder] = list(files_md5_dict.keys())
            self.finished_folders[folder].append(meta_file)
        return

    def include_new_key(self, sample=None):
        self.logsum.feed_key(key=self.current_folder, sample=sample)
        return

    def include_error(self, entry, sample=None):
        self.logsum.add_error(key=self.current_folder, entry=entry, sample=sample)
        return

    def include_warning(self, entry, sample=None):
        self.logsum.add_warning(key=self.current_folder, entry=entry, sample=sample)
        return

    def execute_process(self):
        """Executes different processes depending on the download_option"""
        if not self.relecov_sftp.open_connection():
            self.log.error("Unable to establish connection towards sftp server")
            stderr.print("[red]Unable to establish sftp connection")
            raise ConnectionError("Unable to establish sftp connection")
        target_folders = self.select_target_folders()
        if self.download_option == "delete_only":
            self.log.info("Initiating delete_only process")
            processed_folders = list(target_folders.keys())
            all_folders = []
            for folder in processed_folders:
                subfolders = self.relecov_sftp.list_remote_folders(
                    folder, recursive=True
                )
                all_folders.extend(subfolders)
            all_folders.extend(processed_folders)
            project_folders = {
                folder for folder in processed_folders if folder.count("/") == 1
            }
            all_folders = sorted(
                set(all_folders) - project_folders,
                key=lambda x: x.count("/"),
                reverse=True,
            )
            for folder in all_folders:
                self.current_folder = folder
                self.delete_remote_files(folder)
                self.clean_remote_folder(folder)
                self.log.info(f"Delete process finished in {folder}")
                stderr.print(f"Delete process finished in {folder}")
            for project_folder in project_folders:
                self.delete_remote_files(project_folder)
                self.log.info(f"Cleaned project folder: {project_folder}")
                stderr.print(f"Cleaned project folder: {project_folder}")
        else:
            target_folders, processed_folders = self.merge_subfolders(target_folders)
            self.download(target_folders)

        self.relecov_sftp.close_connection()
        stderr.print(f"Processed {len(processed_folders)} folders: {processed_folders}")
        self.log.info(f"Processed {len(processed_folders)} folders:{processed_folders}")
        if self.logsum.logs:
            self.log.info(
                "Printing process summary to %s", self.platform_storage_folder
            )
            self.parent_create_error_summary(called_module="download")
        else:
            self.log.info("Process log summary was empty. Not generated.")
        processed_folders = list(
            set(os.path.normpath(folder) for folder in processed_folders)
        )
        # If download_option is "download_clean", remove
        # sftp folder content after download is finished
        if self.download_option == "download_clean":
            normal_folders = {
                folder
                for folder in processed_folders
                if "_tmp_processing" not in folder
            }
            for folder in normal_folders:
                if self.relecov_sftp.get_file_list(folder):
                    self.delete_remote_files(folder, skip_seqs=True)
                    self.clean_remote_folder(folder)
            folders_to_clean = copy.deepcopy(self.finished_folders)
            for folder, downloaded_files in folders_to_clean.items():
                self.delete_remote_files(folder, files=downloaded_files)
                self.clean_remote_folder(folder)
                self.log.info(f"Delete process finished in remote {folder}")

            invalid_folders = [
                key for key in target_folders if key not in folders_to_clean
            ]
            for folder in invalid_folders:
                self.rename_remote_folder(folder)
                self.log.info(f"Renamed tmp processing folder: {folder}")

            cleaned_folders = []
            for folder in processed_folders:
                normalized_folder = os.path.normpath(folder)
                if normalized_folder not in cleaned_folders:
                    cleaned_folders.append(normalized_folder)

        self.log.info("Finished download module execution")
        stderr.print("Finished execution")
        return



#!/usr/bin/env python
import rich.console
from jsonschema import Draft202012Validator, FormatChecker
import sys
import os
import openpyxl
from datetime import datetime

import relecov_tools.utils
import relecov_tools.assets.schema_utils.jsonschema_draft
import relecov_tools.assets.schema_utils.custom_validators
from relecov_tools.config_json import ConfigJson
from relecov_tools.base_module import BaseModule


stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class SchemaValidation(BaseModule):
    def __init__(
        self,
        json_file=None,
        json_schema_file=None,
        metadata=None,
        output_dir=None,
        excel_sheet=None,
        registry=None,
    ):
        """Validate json file against the schema"""
        super().__init__(output_dir=output_dir, called_module=__name__)
        config_json = ConfigJson()
        self.log.info("Initiating validation process")
        if json_schema_file is None:
            schema_name = config_json.get_topic_data("json_schemas", "relecov_schema")
            json_schema_file = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "schema", schema_name
            )

        self.json_schema = relecov_tools.utils.read_json_file(json_schema_file)

        if json_file is None:
            json_file = relecov_tools.utils.prompt_path(
                msg="Select the json file to be validated"
            )

        if output_dir is None:
            self.out_folder = relecov_tools.utils.prompt_path(
                msg="Select the folder where excel file with invalid data will be saved"
            )
        else:
            self.out_folder = output_dir

        # Read and check json to validate file
        if not os.path.isfile(json_file):
            stderr.print("[red] Json file does not exist")
            self.log.error("Json file does not exist")
            sys.exit(1)
        self.json_data_file = json_file
        out_path = os.path.dirname(os.path.realpath(self.json_data_file))
        self.lab_code = out_path.split("/")[-2]
        self.logsum = self.parent_log_summary(
            output_dir=self.out_folder, lab_code=self.lab_code, path=out_path
        )

        stderr.print("[blue] Reading the json file")
        self.log.info("Reading the json file")
        self.json_data = relecov_tools.utils.read_json_file(json_file)
        if not isinstance(self.json_data, list):
            stderr.print(f"[red]Invalid json file content in {json_file}.")
            stderr.print("Should be a list of dicts. Create it with read-lab-metadata")
            self.log.error(f"[red]Invalid json file content in {json_file}.")
            self.log.error(
                "Should be a list of dicts. Create it with read-lab-metadata"
            )
            raise TypeError(f"Invalid json file content in {json_file}")
        try:
            batch_id = self.get_batch_id_from_data(self.json_data)
        except ValueError:
            raise ValueError(f"Provided json file {json_file} is empty")
        except AttributeError as e:
            raise ValueError(f"Invalid json file content in {json_file}: {e}")
        self.set_batch_id(batch_id)

        self.metadata = metadata
        try:
            self.sample_id_field = self.get_sample_id_field()
        except ValueError as e:
            self.sample_id_field = None
            self.SAMPLE_FIELD_ERROR = str(e)
        conf_subdata = config_json.get_topic_data("sftp_handle", "metadata_processing")
        if excel_sheet is None:
            try:
                self.excel_sheet = conf_subdata["excel_sheet"]
            except KeyError:
                self.log.error("Default metadata sheet name should be in config file")
                raise
        else:
            self.excel_sheet = excel_sheet

        # Parse file containing sample IDs registry
        if registry is not None and relecov_tools.utils.file_exists(registry):
            self.registry_path = registry
        else:
            config_json = ConfigJson(extra_config=True)
            try:
                default_path = config_json.get_topic_data(
                    "validate_config", "default_sample_id_registry"
                )
            except KeyError:
                default_path = None

            if default_path and relecov_tools.utils.file_exists(default_path):
                self.registry_path = default_path
            else:
                stderr.print(
                    "[yellow]No valid ID registry found. Please select the file manually."
                )
                prompted_path = relecov_tools.utils.prompt_path(
                    "Select the JSON file with registered unique sample IDs"
                )
                if relecov_tools.utils.file_exists(prompted_path):
                    self.registry_path = prompted_path
                else:
                    stderr.print(
                        "[red]No valid ID registry file could be found or selected."
                    )
                    sys.exit(1)

        # Read id registry file
        try:
            self.id_registry = relecov_tools.utils.read_json_file(self.registry_path)
        except Exception as e:
            stderr.print(f"[red]Failed to read ID registry JSON: {e}")
            self.log.error(f"Failed to read ID registry JSON: {e}")
            raise

    def validate_schema(self):
        """Validate json schema against draft"""
        relecov_tools.assets.schema_utils.jsonschema_draft.check_schema_draft(
            self.json_schema, "2020-12"
        )

    def get_sample_id_field(self):
        """Find the name of the field used to track the samples in the given schema"""
        # TODO: Include this field in configuration.json
        sample_id_ontology = "GENEPIO:0000079"
        ontology_match = [
            x
            for x, y in self.json_schema["properties"].items()
            if y.get("ontology") == sample_id_ontology
        ]
        if ontology_match:
            sample_id_field = ontology_match[0]
        else:
            error_text = f"No valid sample ID field ({sample_id_ontology}) in schema"
            raise ValueError(error_text)
        return sample_id_field

    def validate_instances(self):
        """Validate data instances against a validated JSON schema"""

        # Create validator
        validator = Draft202012Validator(
            self.json_schema, format_checker=FormatChecker()
        )
        schema_props = self.json_schema["properties"]

        validated_json_data = []
        invalid_json = []
        errors = {}
        error_keys = {}

        if self.sample_id_field is None:
            log_text = f"Logs keys set to None. Reason: {self.SAMPLE_FIELD_ERROR}"
            self.logsum.add_warning(sample=self.sample_id_field, entry=log_text)

        stderr.print("[blue] Start processing the JSON file")
        self.log.info("Start processing the JSON file")

        # Prepare ID registry (validated samples will be asigned with an unique ID)
        new_ids_to_save = {}

        # Start validation
        for item_row in self.json_data:
            sample_id_value = item_row.get(self.sample_id_field)

            # Collect all errors (don't raise immediately)
            validation_errors = list(validator.iter_errors(item_row))

            # Run the custom validator to check if errors should be ignored
            validation_errors = relecov_tools.assets.schema_utils.custom_validators.validate_with_exceptions(
                self.json_schema, item_row, validation_errors
            )
            if not validation_errors:
                # If sample has validated, then assign an unique id
                if "unique_sample_id" not in item_row:
                    new_id = self.generate_incremental_unique_id(
                        {**self.id_registry, **new_ids_to_save}
                    )
                    item_row["unique_sample_id"] = new_id
                    new_ids_to_save[new_id] = {
                        "sequencing_sample_id": sample_id_value,
                        "lab_code": self.lab_code,
                        "generated_at": datetime.now().isoformat(timespec="seconds"),
                    }
                validated_json_data.append(item_row)
                self.logsum.feed_key(sample=sample_id_value)
            else:
                # Process remaining errors
                for error in validation_errors:
                    try:
                        if error.validator == "required":
                            error_field = list(error.message.split("'"))[1]
                        elif error.validator == "anyOf":
                            missing_fields = []
                            for cond in error.validator_value:
                                if isinstance(cond, dict) and "required" in cond:
                                    missing_fields.extend(cond["required"])
                            error_field = " or ".join(missing_fields)
                        elif error.absolute_path:
                            error_field = str(error.absolute_path[0])
                        else:
                            error_field = error.validator or error.message
                    except Exception as ex:
                        self.log.warning(
                            f"Error extracting error_field from: {error}, {ex}"
                        )
                        error_field = str(error)

                    # Try to get the human-readable label from the schema
                    try:
                        err_field_label = schema_props[error_field]["label"]
                    except KeyError:
                        self.log.error(f"Could not extract label for {error_field}")
                        err_field_label = error_field
                    # Format the error message
                    error.message = error.message.replace(error_field, err_field_label)
                    if error.validator == "pattern" and "date" in error_field:
                        error_text = f"Error in column {err_field_label}. Please provide a date from 2020 to date"
                    elif (
                        error.validator == "format" and error.validator_value == "date"
                    ):  # Modification of default text warning for invalid date format.
                        error_text = f"Error in column {err_field_label}: '{error.instance}' is not a valid date format. Valid format 'YYYY-MM-DD'"
                    else:
                        error_text = (
                            f"Error in column {err_field_label}: {error.message}"
                        )

                    # Log errors for summary
                    error_keys[error.message] = error_field
                    errors[error.message] = errors.get(error.message, 0) + 1
                    self.logsum.add_error(sample=sample_id_value, entry=error_text)

                # Add the invalid row to the list
                invalid_json.append(item_row)

        # For samples that are valid, save its new unique ID in the registry
        if new_ids_to_save:
            self.save_new_ids(new_ids_to_save)

        # Summarize errors
        stderr.print("[blue] --------------------")
        stderr.print("[blue] VALIDATION SUMMARY")
        stderr.print("[blue] --------------------")
        self.log.info("Validation summary:")
        max_length = 250
        for error_type, count in errors.items():
            field_with_error = error_keys[error_type]
            if (
                "is not a 'date'" in error_type
            ):  # Modification of default text warning for invalid date format.
                error_text = f"{count} samples failed validation for {field_with_error}:\n{error_type.split()[0]} is not a valid date format. Valid format 'YYYY-MM-DD"
            else:
                error_text = f"{count} samples failed validation for {field_with_error}:\n{error_type}"
            truncated_msg = (
                error_text[:max_length] + "..."
                if len(error_text) > max_length
                else error_text
            )
            self.logsum.add_warning(entry=truncated_msg)
            stderr.print(f"[red]{truncated_msg}")
            stderr.print("[red] --------------------")

        return validated_json_data, invalid_json

    def create_invalid_metadata(self, invalid_json, metadata, out_folder):
        """Create a new sub excel file having only the samples that were invalid.
        Samples name are checking the Sequencing sample id which are in
        column B (index 1).
        The rows that match the value collected from json file on tag
        collecting_lab_sample_id are removed from excel
        """
        if self.sample_id_field is None:
            log_text = f"Invalid excel file won't be created: {self.SAMPLE_FIELD_ERROR}"
            self.logsum.add_error(entry=log_text)
            return
        self.log.error("Some of the samples in json metadata were not validated")
        stderr.print("[red] Some of the Samples are not validate")
        if metadata is None:
            metadata = relecov_tools.utils.prompt_path(
                msg="Select the metadata file to select those not-validated samples."
            )
        if not os.path.isfile(metadata):
            self.log.error("Metadata file %s does not exist", metadata)
            stderr.print(
                "[red] Unable to create excel file for invalid samples. Metadata file ",
                metadata,
                " does not exist",
            )
            sys.exit(1)
        sample_list = []
        stderr.print("Start preparation of invalid samples")
        self.log.info("Start preparation of invalid samples")
        for row in invalid_json:
            sample_list.append(str(row[self.sample_id_field]))
        wb = openpyxl.load_workbook(metadata)
        try:
            ws_sheet = wb[self.excel_sheet]
        except KeyError:
            logtxt = f"No sheet named {self.excel_sheet} could be found in {metadata}"
            self.log.error(logtxt)
            raise
        tag = "Sample ID given for sequencing"
        # Check if mandatory colum ($tag) is defined in metadata.
        try:
            header_row = [idx + 1 for idx, x in enumerate(ws_sheet.values) if tag in x][
                0
            ]
        except IndexError:
            self.log.error(
                f"Column with tag '{tag}' not found in any row of the Excel sheet."
            )
            stderr.print(f"[red]Column with tag '{tag}' not found. Cannot continue.")
            raise
        row_to_del = []
        row_iterator = ws_sheet.iter_rows(min_row=header_row, max_row=ws_sheet.max_row)
        consec_empty_rows = 0
        id_col = [
            idx for idx, val in enumerate(ws_sheet[header_row]) if val.value == tag
        ][0]
        for row in row_iterator:
            # if no data in 10 consecutive rows, break loop
            if not any(row[x].value for x in range(10)):
                row_to_del.append(row[0].row)
                consec_empty_rows += 1
            if consec_empty_rows > 10:
                break
            consec_empty_rows = 0
            if str(row[id_col].value) not in sample_list:
                row_to_del.append(row[0].row)
        stderr.print("Collected rows to create the excel file")
        if len(row_to_del) > 0:
            row_to_del.sort(reverse=True)
            for idx in row_to_del:
                try:
                    ws_sheet.delete_rows(idx)
                except TypeError as e:
                    self.log.error(
                        "Unable to delete row %s from metadata file because of",
                        idx,
                        e,
                    )
                    stderr.print(f"[red] Unable to delete row {idx} becuase of {e}")
                    sys.exit(1)
        os.makedirs(out_folder, exist_ok=True)
        new_name = "invalid_" + os.path.basename(metadata)
        m_file = os.path.join(out_folder, new_name)
        self.log.info("Saving excel file with the invalid samples")
        stderr.print("Saving excel file with the invalid samples")
        wb.save(m_file)
        return

    def create_validated_json(self, valid_json_data, out_folder):
        """Create a copy of the input json file, keeping only the validated samples

        Args:
            valid_json_data (list(dict)): List of samples metadata as dictionaries
            out_folder (str): path to folder where file will be created
        """
        file_name = "_".join(["validated", os.path.basename(self.json_data_file)])
        file_path = os.path.join(out_folder, file_name)
        self.log.info("Saving Json file with the validated samples in %s", file_path)
        relecov_tools.utils.write_json_to_file(valid_json_data, file_path)
        return

    def generate_incremental_unique_id(self, current_registry, prefix="RLCV"):
        """Generates an incremental unique ID using as baseline the latest record
        in a registry.

        Args:
            current_registry (dict): dict containing sample's unique IDs
            prefix (str, optional): String that preceeds the unique ID. Defaults to "RLCV".

        Returns:
            string: An unique ID
        """
        existing_numbers = []
        for uid in current_registry:
            if uid.startswith(prefix):
                try:
                    number = int(uid.replace(f"{prefix}-", ""))
                    existing_numbers.append(number)
                except ValueError:
                    continue
        next_number = max(existing_numbers, default=0) + 1
        return f"{prefix}-{next_number:09d}"

    def save_new_ids(self, new_ids_dict):
        """Updates sample id registry by adding sample unique ids of new validated samples.

        Args:
            new_ids_dict (dict): Dict of new unique sample IDs.
        """
        updated_registry = {**self.id_registry, **new_ids_dict}
        relecov_tools.utils.write_json_to_file(updated_registry, self.registry_path)
        self.log.info(f"Added {len(new_ids_dict)} new sample IDs to registry")

    def validate(self):
        """Validate samples from metadata, create an excel with invalid samples,
        and a json file with the validated ones
        """
        self.validate_schema()
        valid_json_data, invalid_json = self.validate_instances()
        if not invalid_json:
            stderr.print("[green]Sucessful validation, no invalid file created!!")
            self.log.info("Sucessful validation, no invalid file created.")
        else:
            log_text = "Summary: %s valid and %s invalid samples"
            self.logsum.add_warning(
                entry=log_text % (len(valid_json_data), len(invalid_json))
            )
            self.create_invalid_metadata(invalid_json, self.metadata, self.out_folder)
        if valid_json_data:
            self.create_validated_json(valid_json_data, self.out_folder)
        else:
            log_text = "All the samples were invalid. No valid file created"
            self.logsum.add_error(entry=log_text)
            stderr.print(f"[red]{log_text}")
        self.parent_create_error_summary(called_module="validate")
        return valid_json_data, invalid_json



#!/usr/bin/env python
import json
import os
import yaml
import logging

import relecov_tools.utils

log = logging.getLogger(__name__)


# pass test
class ConfigJson:
    # TODO: Make this path configurable too
    _extra_config_path = os.path.expanduser("~/.relecov_tools/extra_config.json")

    def __init__(
        self,
        json_file=os.path.join(os.path.dirname(__file__), "conf", "configuration.json"),
        extra_config=False,
    ):
        """Load config content in configuration.json and additional config if required

        Args:
            json_file (str, optional): config filepath.
            extra_config (bool, optional): Include content from ~/.relecov_tools/extra_config.json.
        """
        with open(json_file, "r", encoding="utf-8") as fh:
            self.json_data = json.load(fh)

        active_extra_conf = False
        if extra_config:
            if os.path.isfile(ConfigJson._extra_config_path):
                try:
                    with open(
                        ConfigJson._extra_config_path, "r", encoding="utf-8"
                    ) as add_fh:
                        additional_conf = json.load(add_fh)
                    self.json_data.update(additional_conf)
                    active_extra_conf = True
                except (OSError, json.JSONDecodeError) as e:
                    log.warning(
                        f"Could not load extra config: {e}. Using default instead"
                    )
            else:
                log.warning(
                    f"Could not load extra config: {ConfigJson._extra_config_path} does not exist. Using default instead"
                )
                log.warning(
                    "Run ``relecov-tools add-extra-config`` to include additional configuration"
                )
        if not active_extra_conf:
            log.debug("Running with default configuration.")
        else:
            log.debug("Loaded additional configuration.")
        self.topic_config = list(self.json_data.keys())

    def get_configuration(self, topic):
        """Obtain the topic configuration from json data"""
        if topic in self.topic_config:
            return self.json_data[topic]
        return None

    def get_topic_data(self, topic, found):
        """Obtain from topic any forward items from json data"""

        def get_recursive(subtopic, found):
            if isinstance(subtopic, dict):
                for key, val in subtopic.items():
                    if key == found:
                        return val
                    elif isinstance(val, dict):
                        result = get_recursive(val, found)
                        if result is not None:
                            return result
            else:
                return None

        if topic not in self.json_data.keys():
            return None
        if found in self.json_data[topic]:
            return self.json_data[topic][found]
        else:
            return get_recursive(self.json_data[topic], found)

    def include_extra_config(self, config_file, config_name=None, force=False):
        """Include given file content as additional configuration for later usage.

        Args:
            config_name (str): Name of the config key that will be added.
            config_file (str): Path to the input file, either Json or Yaml format.
            force (bool, optional): Force replacement of existing configuration if needed.

        Raises:
            ValueError: If provided config_file does not have a supported extension.
        """

        def validate_new_config(config_name, current_conf, force=False):
            """Check if config_name is already in configuration"""
            if config_name in current_conf.keys():
                if force:
                    log.info(
                        f"`{config_name}` already in config. Replacing its content..."
                    )
                    return True
                else:
                    err_txt = f"Cannot add `{config_name}`: already in config. Set `force` to force replacement"
                    log.error(err_txt)
                    return False
            else:
                return True

        def rec_merge_config(current_conf, new_conf, force=False):
            """Recursively add new configuration without deleting the existing keys"""
            for k, v in new_conf.items():
                if isinstance(v, dict) and k in current_conf:
                    rec_merge_config(current_conf[k], v, force=force)
                else:
                    change_msg = f"{k}: {current_conf.get(k, '')} -> {v}"
                    if not validate_new_config(k, current_conf, force=force):
                        summary["Canceled"].append(change_msg)
                        continue
                    summary["Included"].append(change_msg)
                    current_conf[k] = v

        if not os.path.isfile(str(config_file)):
            raise FileNotFoundError(f"Extra config file {config_file} does not exist")
        if os.path.isfile(ConfigJson._extra_config_path):
            with open(ConfigJson._extra_config_path, "r", encoding="utf-8") as f:
                additional_config = json.load(f)
        else:
            additional_config = None

        summary = {"Included": [], "Canceled": []}
        valid_exts = (".json", ".yaml", ".yml")
        os.makedirs(os.path.dirname(ConfigJson._extra_config_path), exist_ok=True)
        with open(config_file, "r") as fh:
            if config_file.endswith(".json"):
                file_content = json.load(fh)
            elif config_file.endswith((".yaml", ".yml")):
                file_content = yaml.load(fh, Loader=yaml.FullLoader)
            else:
                raise ValueError(
                    f"config_file {config_file} extension is not supported. Use: {valid_exts}"
                )
        if additional_config is not None:
            if config_name is None:
                rec_merge_config(additional_config, file_content, force=force)
            else:
                if validate_new_config(config_name, self.json_data, force=force):
                    msg = f"{config_name}: {file_content}"
                    summary["Included"].append(msg)
                    additional_config.update({config_name: file_content})
        else:
            if config_name is None:
                additional_config = file_content
            else:
                additional_config = {config_name: file_content}
            summary["Included"].extend([k for k in additional_config.keys()])
        relecov_tools.utils.write_json_to_file(
            additional_config, ConfigJson._extra_config_path
        )
        log.info("Finished including extra configuration")
        print("Update summary:")
        for state, changes in summary.items():
            print(state, ":\n", "\n".join([str(msg) for msg in changes]))
        return

    def remove_extra_config(self, config_name):
        """Remove key from extra_config configuration file

        Args:
            config_name (str): _description_
        """
        if config_name is None:
            try:
                os.remove(ConfigJson._extra_config_path)
                log.info("Removed extra config file.")
            except OSError as e:
                log.error(f"Could not remove extra config file: {e}")
                return
        else:
            with open(ConfigJson._extra_config_path, "r") as fh:
                additional_config = json.load(fh)
            try:
                additional_config.pop(config_name)
            except KeyError:
                log.error(f"{config_name} not found in extra_config for removal")
                return
            except OSError as e:
                log.error(f"Could not remove {config_name} key: {e}")
                return
            relecov_tools.utils.write_json_to_file(
                additional_config, ConfigJson._extra_config_path
            )
            log.info(f"Successfully removed {config_name} from extra config")
        print(f"Finished clearing extra config: {config_name}")
        return

    def get_lab_code(self, submitting_institution):
        """Get the corresponding code for the given submitting institution"""
        if submitting_institution is None:
            log.warning("No submitting institution could be found to update lab_code")
            return
        institutions_config = self.get_configuration("institutions_config")
        if not institutions_config:
            log.warning("No institutions_config found. Could not extract lab_code")
            return
        for code, conf in institutions_config.items():
            if submitting_institution in conf.get("institution_name"):
                return code
        else:
            log.warning(f"{submitting_institution} not found in institutions_config")
        return



#!/usr/bin/env python
import logging
from collections import OrderedDict

log = logging.getLogger(__name__)


class PhagePlusSchema:
    def __init__(self, schema):
        self.schema = schema
        self.ontology = {}
        for key, values in schema["properties"].items():
            self.ontology[values["ontology"]] = key
        self.properties = list(schema["properties"].keys())

    def get_gontology(self, property_item):
        """Return the geontology value for a property in the schema"""
        try:
            return self.schema["properties"][property_item]["ontology"]
        except KeyError as e:
            log.error("geontology value %s %s", property_item, e)
            return None

    def maping_schemas_based_on_geontology(self, mapped_to_schema):
        """Return a dictionnary with the properties of the mapped_to_schema
        as key and properties of phagePlusSchema as value
        """
        mapped_dict = OrderedDict()
        for key, values in mapped_to_schema["properties"].items():
            try:
                mapped_dict[key] = self.ontology[values["ontology"]]
            except KeyError as e:
                log.error("Enable to map schema, because of %s is not defined", e)
                # There is no exact match on ontology. Search for the parent
                # to be implemented later
                pass
        return mapped_dict

    def get_schema_properties(self):
        """Return the properties defined in the schema"""
        return self.properties
